"""
Usage
-----
Collection objects repesents a single Rockset collection.
These objects are generally created using a Rockset Client_
object using methods such as::

    from rockset import Client

    # connect to Rockset
    rs = Client()

    # create a new collection
    user_events = rs.Collection.create('user-events')

    # retrieve an existing collection
    users = rs.Collection.retrieve('users')

You can add documents to the collection using the ``add_docs()`` method. Each
document in a collection is uniquely identified by its ``_id`` field.

If documents added have ``_id`` fields that match existing documents,
then their contents will be merged. Otherwise, the new documents will be
added to the collection.

You can remove documents from a collection using the ``remove_docs()`` method.

Refer to the Query_ module for documentation and examples on how to query
collections.

Example
-------
::

    from rockset import Client, Q, F

    # connect securely to Rockset
    rs = Client()

    # retrieve the relevant collection
    emails = rs.Collection.retrieve('emails')

    # look for all emails to johndoe that contains the term 'secret'
    johndoe_secret_q = Q('emails').where(
        (F["to"].startswith('johndoe@')) & (F["body"][:] == 'secret')
    )

    # query the collection
    docs = rs.sql(johndoe_secret_q).results()

.. _Collection.create:

Create a new collection
-----------------------
Creating a collection using the Client_ object is as simple as
calling ``client.Collection.create("my-new-collection")``::

    from rockset import Client
    rs = Client()
    new_collection = rs.Collection.create("my-new-collection")

    # create a collection in a workspace
    leads_collection = rs.Collection.create("leads",
                                                 workspace="marketing")

    # create a collection and map timestamp field to event-time
    field_mappings = [
      rs.FieldMapping.mapping(
        name="transformation1",
        input_fields=[
          rs.FieldMapping.input_field(
            field_name="ts",
              if_missing="PASS",
              is_drop=True,
              param="ts"
          )
        ],
        output_field=rs.FieldMapping.output_field(
            field_name="_event_time",
            sql_expression="CAST(:ts AS TIMESTAMP)",
            on_error="SKIP"
        )
      )
    ]

    event_data_collection = rs.Collection.create("event-data-collection",
                                                 field_mappings=field_mappings)

Creating a collection with a retention duration of 10 days::

    from rockset import Client
    rs = Client()
    new_collection_with_retention = rs.Collection.create("my-event-collection",
                                                    retention_secs=10*24*60*60)

.. _Collection.list:

List all collections
--------------------
List all collections using the Client_ object using::

    from rockset import Client
    rs = Client()

    # list all collections
    collections = rs.Collection.list()

.. _Collection.retrieve:

Retrieve an existing collection
-------------------------------
Retrive a collection to run various operations on that collection
such as adding or removing documents or executing queries::

    from rockset import Client
    rs = Client()
    users = rs.Collection.retrieve('users')

    # retrieve a collection in a workspace
    users = rs.Collection.retrieve('users', workspace='marketing')


.. _Collection.describe:

Describe an existing collection
-------------------------------
The ``describe`` method can be used to fetch all the details about the collection
such as what data sets act as the collection's sources, various performance and
usage statistics::

    from rockset import Client
    rs = Client()
    users = rs.Collection.retrieve('users')
    print(users.describe())

.. _Collection.drop:

Drop a collection
-----------------
Use the ``drop()`` method to remove a collection permanently from Rockset.

.. note:: This is a permanent and non-recoverable operation. Beware.

::

    from rockset import Client
    rs = Client()
    users = rs.Collection.retrieve('users')
    users.drop()

.. _Collection.add_docs:

Add documents to a collection
-----------------------------
Python dicts can be added as documents to a collection using the ``add_docs``
method. Documents are uniquely identified by the ``_id`` field. If an input
document does not have an ``_id`` field, then an unique id will be assigned
by Rockset.

If the ``_id`` field of an input document does not match an existing document,
then a new document will be created.

If the ``_id`` field of an input document matches an existing document,
then the new document will be merged with the existing document::

    from rockset import Client
    import json

    rs = Client()
    with open('my-json-array-of-dicts.json') as data_fh:
        ret = rs.Collection.add_docs('users', json.load(data_fh))

.. _Collection.remove_docs:

Delete documents from a collection
----------------------------------
Remove documents from a collection using the ``remove_docs`` method::

    from rockset import Client

    rs = Client()
    users_to_remove = ['user007', 'user042', 'user435']
    docs_to_remove = [{'_id': u} for u in users_to_remove]
    ret = rs.Collection.remove_docs('users', docs_to_remove)

"""
from .exception import InputError
from .resource import Resource

from rockset.swagger_client.api import (CollectionsApi, DocumentsApi)
from rockset.swagger_client.models import (
    CreateCollectionRequest, AddDocumentsRequest, PatchDocumentsRequest,
    DeleteDocumentsRequest
)


class Collection(Resource):
    """
    Collection objects represent a single Rockset collection.

    Objects of this class are never instantiated directly and are
    generally returned by methods such as::

        from rockset import Client
        rs = Client()
        first = rs.Collection.create('my-first-collection')
        another = rs.Collection.retrieve('another-collection')

    See more examples and documentation here.
    """
    @classmethod
    def create(
        cls,
        name,
        workspace="commons",
        description=None,
        sources=None,
        retention_secs=None,
        field_mappings=None,
        clustering_key=None,
        **kwargs
    ):
        """Creates a new Rockset collection.

        Use it via rockset.Client().Collection.create()

        Only alphanumeric characters, ``_``, and ``-`` are allowed
        in collection names.

        Args:
            name (str): name of the collection to be created.
            description (str): a human readable description of the collection
            sources (Source): array of Source objects that defines the set
                of input data sources for this collection
            retention_secs (int): an integer representing minimum duration (in seconds),
                for which documents are retained in this collection before being automatically deleted.
            field_mappings (FieldMapping): array of FieldMapping objects that
                defines the set of transformations to apply on all documents
            clustering_key (ClusteringKey): array of ClusteringKey objects
                that defines the clustering config for this collection
        Returns:
            Collection: Collection object
        """
        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage. '
                'use rockset.Client().Collection.create() instead.'
            )
        client = kwargs.pop('client')

        kwargs['description'] = description
        kwargs['sources'] = sources
        kwargs['retention_secs'] = retention_secs
        kwargs['field_mappings'] = field_mappings
        kwargs['clustering_key'] = clustering_key

        req = CreateCollectionRequest(name=name, **kwargs)
        collection = CollectionsApi(client).create(
            workspace=workspace, body=req
        ).get('data').to_dict()

        return cls(client=client, **collection)

    @classmethod
    def retrieve(cls, name, **kwargs):
        """Retrieves details of a single collection

        Use it via rockset.Client().Collection.retrieve()

        Args:
            name (str): Name of the collection

        Returns:
            Collection: Collection object
        """
        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage. '
                'use rockset.Client().Collection.create() instead.'
            )
        c = cls(name=name, **kwargs)
        c.describe()

        return c

    @classmethod
    def list(cls, **kwargs):
        """Returns list of all collections.

        Use it via rockset.Client().Collection.list()

        Returns:
            List: A list of Collection objects
        """
        if 'client' not in kwargs:
            raise ValueError(
                'incorrect API usage. '
                'use rockset.Client().Collection.list() instead.'
            )
        client = kwargs.pop('client')
        workspace = kwargs.pop('workspace', None)
        capi = CollectionsApi(client)

        if workspace:
            collections = capi.workspace(workspace).get('data')
        else:
            collections = capi.list().get('data')

        ret = []
        for c in collections:
            if isinstance(c, dict):
                c_dict = c
            else:
                c_dict = c.to_dict()

            ret.append(cls(client=client, **c_dict))

        return ret

    def __init__(self, *args, **kwargs):
        kwargs['type'] = 'COLLECTION'
        super(Collection, self).__init__(*args, **kwargs)
        self.docs_per_call = 1000

    def _chopper(self, docs):
        return [
            docs[i:i + self.docs_per_call]
            for i in range(0, len(docs), self.docs_per_call)
        ]

    def _validate_doclist(self, docs, action):
        if type(docs) != list:
            raise InputError(message='arg "docs" is not a list of dicts')
        for doc in docs:
            if not isinstance(doc, dict):
                raise InputError(
                    message=
                    f'cannot {action} a document that is not a dict or rockset.Document'
                )

    # instance methods
    def describe(self):
        """Returns all properties of the collection as a dict.

        Returns:
            dict: properties of the collection
        """
        return super(Collection, self).describe()

    def drop(self):
        """Deletes the collection represented by this object.

        If successful, the current object will contain
        a property named ``dropped`` with value ``True``

        Example::

            ...
            print(my_coll.asdict())
            my_coll.drop()
            print(my_coll.dropped)       # will print True
            ...
        """
        super(Collection, self).drop()
        return

    def query(self, q, timeout=None, flood_all_leaves=False):
        return super(Collection, self).query(
            q, timeout=timeout, flood_all_leaves=flood_all_leaves
        )

    def setstate(self, newstate):
        return super(Collection, self).setstate(newstate)

    def add_docs(self, docs, timeout=None):
        """Adds or merges documents to the collection. Provides document
        level atomicity.

        Documents within a collection are uniquely identified by the
        ``_id`` field. If input document does not specify ``_id``, then
        an unique UUID will be assigned to the document.

        If the ``_id`` field of an input document does not match with
        any existing collection documents, then the input document will
        be inserted.

        If the ``_id`` field of an input document matches with an
        existing collection document, then the input document will be
        merged atomically as described below:

        * All fields present in both the input document and the collection
          document will be updated to values from the input document.
        * Fields present in the input document but not the collection
          document will be inserted.
        * Fields present in the collection document but not the input
          document will be left untouched.

        All fields within every input document will be inserted or updated
        atomically. No atomicity guarantees are provided across two different
        documents added.

        Example::

            from rockset import Client

            rs = Client()
            docs = [
                {"_id": "u42", "name": {"first": "Jim", "last": "Gray"}},
                {"_id": "u1201", "name": {"first": "David", "last": "DeWitt"}},
            ]
            results = rs.Collection.add_docs("my-favorite-collection", docs)
            ...

        Args:
            name (str): Name of the collection
            docs (list of dicts): New documents to be added or merged
            timeout (int): Client side timeout. When specified,
                RequestTimeout_ exception will
                be thrown upon timeout expiration. By default, the client
                will wait indefinitely until it receives results or an
                error from the server.

        Returns:
            Dict: The response dict will have 1 field: ``data``

            The ``data`` field will be a list of document status records,
            one for each input document indexed in the same order as the list
            of input documents provided as part of the request. Each of those
            document status records will have fields such as the document
            ``_id``, ``_collection`` name, ``status`` describing if that
            particular document add request succeeded or not, and an optional
            ``error`` field with more details.
        """
        self._validate_doclist(docs, 'add')

        # chunk docs to operate in batches
        retval = []
        for chunk in self._chopper(docs):
            request = AddDocumentsRequest(data=chunk)
            retval.extend(
                DocumentsApi(self.client).add(
                    workspace=self.workspace,
                    collection=self.name,
                    body=request,
                    _request_timeout=timeout
                ).get('data')
            )
        return retval

    def patch_docs(self, docpatches, timeout=None):
        """Updates documents in the collection.

        This method expects a list of docpatches, where each docpatch
        describes a set of updates that need to be applied to a single
        document in the collection.

        All updates specified in a single docpatch will be applied atomically
        to the document. If a single patch operation specified in a patch
        fails, the entire patch operation will be aborted.

        Each docpatch is a dict that is required to have 2 fields\:

        1. ``_id`` that holds the _id field of the document to be updated

        2. ``patch`` that holds a list of patch operations to be applied
           to that document, following the JSONPatch standard defined at
           http://jsonpatch.com

        Example::

            from rockset import Client
            rs = Client()

            docpatch = {
                "_id": "u42",
                "patch": [
                    {"op": "replace", "path": "/name/middle", "value": "Nicholas"}
                ]
            }
            rs.Collection.patch_docs('my-collection', [docpatch])

        Each patch is a list of individual patch operations, where each patch
        operation specifies how a particular field or field path within the
        target document should be updated.

        Each patch operation is a dict with a key called "op" (for
        operation) and few more keys that act as arguments to the "op", which
        differ from one operation type to another. The JSONPatch standard
        defines several types of patch operations, their arguments and their
        behavior. Refer to http://jsonpatch.com/#operations for more details.

        If a patch opertion's argument is a field path, then it is specified
        using the JSON Pointer standard defined at https://tools.ietf.org/html/rfc6901
        In essence, field paths are represented as a string of tokens
        separated by ``/`` characters. These tokens either specify keys in
        objects or indexes into arrays, and arrays are 0-based.

        For example, in this document::

            {
                "biscuits": [
                    { "name": "Digestive" },
                    { "name": "Choco Leibniz" }
                ]
            }

            "/biscuits" would point to the array of biscuits
            "/biscuits/1/name" would point to "Choco Leibniz".


        Here are some examples of individual patch operations:

        * Add

          Example::

              { "op": "add", "path": "/biscuits/1", "value": { "name": "Ginger Nut" } }

          Adds a value to an object or inserts it into an array. In the
          case of an array, the value is inserted before the given index.
          The ``-`` character can be used instead of an index to insert at the
          end of an array.

        * Remove

          Example::

              { "op": "remove", "path": "/biscuits" }

          Removes a value from an object or array.

          Another Example::

              { "op": "remove", "path": "/biscuits/0" }

          Removes the first element of the array at biscuits
          (or just removes the "0" key if biscuits is an object)

        * Replace

          Example::

              { "op": "replace", "path": "/biscuits/0/name", "value": "Chocolate Digestive" }

          Replaces a value. Equivalent to a "remove" followed by an "add".

        * Copy

          Example::

              { "op": "copy", "from": "/biscuits/0", "path": "/best_biscuit" }

          Copies a value from one location to another within the JSON document.
          Both "from" and "path" are JSON Pointers.

        * Move

          Example::

              { "op": "move", "from": "/biscuits", "path": "/cookies" }

          Moves a value from one location to the other. Both "from" and "path" are
          JSON Pointers.

        * Test

          Example::

              { "op": "test", "path": "/best_biscuit/name", "value": "Choco Leibniz" }

          Tests that the specified value is set in the document. If the test
          fails, then the patch as a whole will not apply.

        Args:
            name (str): Name of the collection
            docpatches (list of dicts): List of document patches to be applied.
            timeout (int): Client side timeout. When specified,
                RequestTimeout_ exception will
                be thrown upon timeout expiration. By default, the client
                will wait indefinitely until it receives results or an
                error from the server.

        Returns:
            Dict: The response dict will have 1 field: ``data``.

            The ``data`` field will be a list of document status records,
            one for each input document indexed in the same order as the list
            of input documents provided as part of the request. Each of those
            document status records will have fields such as the document
            ``_id``, ``_collection`` name, ``status`` describing if that
            particular document add request succeeded or not, and an optional
            ``error`` field with more details.
        """
        self._validate_doclist(docpatches, 'patch')

        # chunk docs to operate in batches
        retval = []
        for chunk in self._chopper(docpatches):
            request = PatchDocumentsRequest(data=chunk)
            retval.extend(
                DocumentsApi(self.client).patch(
                    workspace=self.workspace,
                    collection=self.name,
                    body=request,
                    _request_timeout=timeout
                ).get('data')
            )
        return retval

    def remove_docs(self, docs, timeout=None):
        """Deletes documents from the collection. The ``_id`` field needs to
        be populated in each input document. Other fields in each document
        will be ignored.

        Args:
            name (str): Name of the collection
            docs (list of dicts): Documents to be deleted.
            timeout (int): Client side timeout. When specified,
                RequestTimeout_ exception will
                be thrown upon timeout expiration. By default, the client
                will wait indefinitely until it receives results or an
                error from the server.

        Returns:
            Dict: The response dict will have 1 field: ``data``.

            The ``data`` field will be a list of document status records,
            one for each input document indexed in the same order as the list
            of input documents provided as part of the request. Each of those
            document status records will have fields such as the document
            ``_id``, ``_collection`` name, ``status`` describing if that
            particular document add request succeeded or not, and an optional
            ``error`` field with more details.
        """
        self._validate_doclist(docs, 'remove')

        # chunk docs to operate in batches
        retval = []
        docids = [{'_id': doc.get('_id', None)} for doc in docs]
        for chunk in self._chopper(docids):
            request = DeleteDocumentsRequest(data=chunk)
            retval.extend(
                DocumentsApi(self.client).delete(
                    workspace=self.workspace,
                    collection=self.name,
                    body=request,
                    _request_timeout=timeout
                ).get('data')
            )
        return retval


__all__ = [
    'Collection',
]



from .command_auth import AuthCommand
from .util.parse_util import parse_collection_path


class Remove(AuthCommand):
    def usage(self):
        return """
usage: rock rm [-h] <name> <document_id>...

Remove one or more documents from a Rockset collection.

arguments:
    <name>              name of the collection
    <document_id>       id of the documents you wish to remove from the
                        collection

options:
    -h, --help          show this help message and exit
        """

    def go(self):
        workspace, name = parse_collection_path(self.name)
        resource = self.client.retrieve(name, workspace=workspace)
        docs = [{'_id': docid} for docid in self.document_id]
        self.print_list(
            0,
            resource.remove_docs(docs=docs),
            field_order=['collection', 'id', 'status', 'error']
        )
        return 0



# coding: utf-8

"""
    REST API

    Rockset's REST API allows for creating and managing all resources in Rockset. Each supported endpoint is documented below.  All requests must be authorized with a Rockset API key, which can be created in the [Rockset console](https://console.rockset.com). The API key must be provided as `ApiKey <api_key>` in the `Authorization` request header. For example: ``` Authorization: ApiKey aB35kDjg93J5nsf4GjwMeErAVd832F7ad4vhsW1S02kfZiab42sTsfW5Sxt25asT ```  All endpoints are only accessible via https.  Build something awesome!  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class Alias(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'description': 'str',
        'workspace': 'str',
        'creator_email': 'str',
        'collections': 'list[str]',
        'state': 'str',
        'created_at': 'str',
        'modified_at': 'str'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'workspace': 'workspace',
        'creator_email': 'creator_email',
        'collections': 'collections',
        'state': 'state',
        'created_at': 'created_at',
        'modified_at': 'modified_at'
    }

    def __init__(self, **kwargs):  # noqa: E501
        """Alias - a model defined in Swagger"""  # noqa: E501

        self._name = None
        self._description = None
        self._workspace = None
        self._creator_email = None
        self._collections = None
        self._state = None
        self._created_at = None
        self._modified_at = None
        self.discriminator = None

        self.name = kwargs.pop('name', None)
        self.description = kwargs.pop('description', None)
        self.workspace = kwargs.pop('workspace', None)
        self.creator_email = kwargs.pop('creator_email', None)
        self.collections = kwargs.pop('collections', None)
        self.state = kwargs.pop('state', None)
        self.created_at = kwargs.pop('created_at', None)
        self.modified_at = kwargs.pop('modified_at', None)

    @property
    def name(self):
        """Gets the name of this Alias.  # noqa: E501

        name of the alias  # noqa: E501

        :return: The name of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Alias.

        name of the alias  # noqa: E501

        :param name: The name of this Alias.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def description(self):
        """Gets the description of this Alias.  # noqa: E501

        alias description  # noqa: E501

        :return: The description of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this Alias.

        alias description  # noqa: E501

        :param description: The description of this Alias.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def workspace(self):
        """Gets the workspace of this Alias.  # noqa: E501

        name of the workspace  # noqa: E501

        :return: The workspace of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):
        """Sets the workspace of this Alias.

        name of the workspace  # noqa: E501

        :param workspace: The workspace of this Alias.  # noqa: E501
        :type: str
        """

        self._workspace = workspace

    @property
    def creator_email(self):
        """Gets the creator_email of this Alias.  # noqa: E501

        email of the creator  # noqa: E501

        :return: The creator_email of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this Alias.

        email of the creator  # noqa: E501

        :param creator_email: The creator_email of this Alias.  # noqa: E501
        :type: str
        """

        self._creator_email = creator_email

    @property
    def collections(self):
        """Gets the collections of this Alias.  # noqa: E501

        list of fully qualified collection names referenced by alias  # noqa: E501

        :return: The collections of this Alias.  # noqa: E501
        :rtype: list[str]
        """
        return self._collections

    @collections.setter
    def collections(self, collections):
        """Sets the collections of this Alias.

        list of fully qualified collection names referenced by alias  # noqa: E501

        :param collections: The collections of this Alias.  # noqa: E501
        :type: list[str]
        """

        self._collections = collections

    @property
    def state(self):
        """Gets the state of this Alias.  # noqa: E501

        state of the alias  # noqa: E501

        :return: The state of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this Alias.

        state of the alias  # noqa: E501

        :param state: The state of this Alias.  # noqa: E501
        :type: str
        """

        self._state = state

    @property
    def created_at(self):
        """Gets the created_at of this Alias.  # noqa: E501

        ISO-8601 date  # noqa: E501

        :return: The created_at of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Alias.

        ISO-8601 date  # noqa: E501

        :param created_at: The created_at of this Alias.  # noqa: E501
        :type: str
        """

        self._created_at = created_at

    @property
    def modified_at(self):
        """Gets the modified_at of this Alias.  # noqa: E501

        ISO-8601 date  # noqa: E501

        :return: The modified_at of this Alias.  # noqa: E501
        :rtype: str
        """
        return self._modified_at

    @modified_at.setter
    def modified_at(self, modified_at):
        """Sets the modified_at of this Alias.

        ISO-8601 date  # noqa: E501

        :param modified_at: The modified_at of this Alias.  # noqa: E501
        :type: str
        """

        self._modified_at = modified_at

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Alias, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Alias):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, item):
        return getattr(self, item)

    def items(self):
        return self.to_dict().items()

    def __setitem__(self, item, value):
        return seattr(self, item, value)



import re
import sys
import unittest
from argparse import ArgumentParser


def main():
    ap = ArgumentParser(description='Rockset unittest main')
    ap.add_argument(
        '-b', '--buffer', help='Buffer output during run', action='store_true'
    )
    ap.add_argument(
        '-c', '--catch', help='Catch Ctrl-C during run', action='store_true'
    )
    ap.add_argument(
        '-f', '--failfast', help='Stop on first failure', action='store_true'
    )
    ap.add_argument(
        '--locals',
        help='Show local variables in tracebacks',
        action='store_true'
    )
    ap.add_argument(
        '-v', '--verbose', help='Verbose output', action='store_true'
    )
    ap.add_argument(
        '-s', '--start-directory', help='Start directory', default='.'
    )
    ap.add_argument('patterns', help='Test regexes to run', nargs='*')

    args = ap.parse_args()
    patterns = [re.compile(p) for p in args.patterns]

    def filter_tests(tests):
        accepted = []
        for test in tests:
            if isinstance(test, unittest.TestSuite):
                accepted.append(test)
                continue
            test_id = test.id()
            if (not patterns) or any(p.match(test_id) for p in patterns):
                accepted.append(test)

        return unittest.TestSuite(accepted)

    if args.catch:
        unittest.installHandler()

    loader = unittest.TestLoader()
    loader.suiteClass = filter_tests
    suite = loader.discover(args.start_directory)
    runner = unittest.TextTestRunner(
        verbosity=2 if args.verbose else 1,
        failfast=args.failfast,
        buffer=args.buffer
    )
    result = runner.run(suite)
    if not result.wasSuccessful():
        sys.exit(1)


if __name__ == '__main__':
    main()



# coding: utf-8

"""
    REST API

    Rockset's REST API allows for creating and managing all resources in Rockset. Each supported endpoint is documented below.  All requests must be authorized with a Rockset API key, which can be created in the [Rockset console](https://console.rockset.com). The API key must be provided as `ApiKey <api_key>` in the `Authorization` request header. For example: ``` Authorization: ApiKey aB35kDjg93J5nsf4GjwMeErAVd832F7ad4vhsW1S02kfZiab42sTsfW5Sxt25asT ```  All endpoints are only accessible via https.  Build something awesome!  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class ApiKey(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'created_at': 'str',
        'name': 'str',
        'key': 'str',
        'last_access_time': 'str',
        'role': 'str',
        'created_by': 'str'
    }

    attribute_map = {
        'created_at': 'created_at',
        'name': 'name',
        'key': 'key',
        'last_access_time': 'last_access_time',
        'role': 'role',
        'created_by': 'created_by'
    }

    def __init__(self, name, key, **kwargs):  # noqa: E501
        """ApiKey - a model defined in Swagger"""  # noqa: E501

        self._created_at = None
        self._name = None
        self._key = None
        self._last_access_time = None
        self._role = None
        self._created_by = None
        self.discriminator = None

        self.created_at = kwargs.pop('created_at', None)
        self.name = name
        self.key = key
        self.last_access_time = kwargs.pop('last_access_time', None)
        self.role = kwargs.pop('role', None)
        self.created_by = kwargs.pop('created_by', None)

    @property
    def created_at(self):
        """Gets the created_at of this ApiKey.  # noqa: E501

        ISO-8601 date  # noqa: E501

        :return: The created_at of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this ApiKey.

        ISO-8601 date  # noqa: E501

        :param created_at: The created_at of this ApiKey.  # noqa: E501
        :type: str
        """

        self._created_at = created_at

    @property
    def name(self):
        """Gets the name of this ApiKey.  # noqa: E501

        descriptive label  # noqa: E501

        :return: The name of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ApiKey.

        descriptive label  # noqa: E501

        :param name: The name of this ApiKey.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def key(self):
        """Gets the key of this ApiKey.  # noqa: E501

        string of 64 alphanumeric characters.  # noqa: E501

        :return: The key of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key):
        """Sets the key of this ApiKey.

        string of 64 alphanumeric characters.  # noqa: E501

        :param key: The key of this ApiKey.  # noqa: E501
        :type: str
        """

        self._key = key

    @property
    def last_access_time(self):
        """Gets the last_access_time of this ApiKey.  # noqa: E501

        ISO-8601 date  # noqa: E501

        :return: The last_access_time of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._last_access_time

    @last_access_time.setter
    def last_access_time(self, last_access_time):
        """Sets the last_access_time of this ApiKey.

        ISO-8601 date  # noqa: E501

        :param last_access_time: The last_access_time of this ApiKey.  # noqa: E501
        :type: str
        """

        self._last_access_time = last_access_time

    @property
    def role(self):
        """Gets the role of this ApiKey.  # noqa: E501


        :return: The role of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._role

    @role.setter
    def role(self, role):
        """Sets the role of this ApiKey.


        :param role: The role of this ApiKey.  # noqa: E501
        :type: str
        """

        self._role = role

    @property
    def created_by(self):
        """Gets the created_by of this ApiKey.  # noqa: E501


        :return: The created_by of this ApiKey.  # noqa: E501
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ApiKey.


        :param created_by: The created_by of this ApiKey.  # noqa: E501
        :type: str
        """

        self._created_by = created_by

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ApiKey, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ApiKey):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, item):
        return getattr(self, item)

    def items(self):
        return self.to_dict().items()

    def __setitem__(self, item, value):
        return seattr(self, item, value)



def parse_collection_path(path):
    components = path.split(".")
    name = components[-1]
    if len(components) == 1:
        workspace = 'commons'
    else:
        workspace = '.'.join(components[:-1])
    return workspace, name



import json
import os
import rockset
import stat

VERSION_KEY = 'version'
VERSION_VALUE = 'v1'
AUTH_TOKEN_KEY = 'auth'
ACTIVE_PROFILE_KEY = 'active_profile'
PROFILES_KEY = 'profiles'


class Credentials(object):
    def __init__(self):
        # init config_dir to be `homedir`/.rockset
        self.config_dir = rockset.Client.config_dir()
        # init creds_file to be `config_dir`/.rockset/credentials
        self.creds_file = os.path.join(self.config_dir, 'credentials')

        # init creds
        self.creds = {
            VERSION_KEY: None,
            AUTH_TOKEN_KEY: None,
            ACTIVE_PROFILE_KEY: None,
            PROFILES_KEY: {},
        }
        self.read()
        return

    def setup(self):
        # Create config_dir if it doesn't exist
        if not os.path.isdir(self.config_dir):
            os.makedirs(self.config_dir)

        # bail if file already exists and has the correct version
        if self.creds[VERSION_KEY] == VERSION_VALUE:
            return

        # Create credentials_file or truncate it if exists
        # it either doesn't exist or is from previous version
        with os.fdopen(
            os.open(self.creds_file, os.O_WRONLY | os.O_CREAT, 0o600), 'w'
        ) as _:
            os.chmod(self.creds_file, stat.S_IRUSR | stat.S_IWUSR)

        return

    def read(self):
        # init all available credentials
        if (
            os.path.exists(self.creds_file) and
            os.path.getsize(self.creds_file) > 0
        ):
            with open(self.creds_file, 'r') as cfh:
                self.creds.update(json.load(cfh))
            self.migrate()
        return

    def migrate(self):
        modified = False
        # go through all api_server entries and rstrip('/')
        # implement all migration logic here when we bump VERSION_VALUE
        for profile in self.creds[PROFILES_KEY]:
            profile_details = self.creds[PROFILES_KEY].get(profile, {})
            api_server = profile_details.get('api_server', '')
            if api_server != api_server.rstrip('/'):
                profile_details.update({'api_server': api_server.rstrip('/')})
                modified = True
        if modified:
            self.save()
        return modified

    def save(self):
        with open(self.creds_file, 'w') as cfh:
            self.creds[VERSION_KEY] = VERSION_VALUE
            json.dump(self.creds, cfh, indent=4)
        return self.creds_file

    def active_profile(self, profile=None):
        if profile is not None and profile in self.creds[PROFILES_KEY]:
            self.creds[ACTIVE_PROFILE_KEY] = profile
        return self.creds[ACTIVE_PROFILE_KEY]

    def auth(self, user=None, auth_token=None):
        if auth_token is not None:
            if user is not None:
                auth_token.update({'user': user})
            self.creds[AUTH_TOKEN_KEY] = auth_token
        return self.creds[AUTH_TOKEN_KEY]

    def set(self, profile=None, api_key=None, api_server=None, auth_token=None):
        if profile is None:
            return None
        profile_details = self.creds[PROFILES_KEY].get(profile, {})
        if api_key is not None:
            profile_details.update({'api_key': api_key})
        if api_server is not None and len(api_server) > 0:
            if api_server[-1] == '/':
                api_server = api_server[:-1]
            profile_details.update({'api_server': api_server})
        if auth_token is not None:
            profile_details.update({'auth': auth_token})
        self.creds[PROFILES_KEY][profile] = profile_details
        return profile_details

    def get(self, profile=None):
        if profile is None:
            profile = self.active_profile()
        return self.creds[PROFILES_KEY].get(profile, {})

    def delete(self, profile):
        if self.creds[ACTIVE_PROFILE_KEY] == profile:
            self.creds[ACTIVE_PROFILE_KEY] = None
        return self.creds[PROFILES_KEY].pop(profile, None)

    def get_all_profiles(self):
        return self.creds[PROFILES_KEY]



"""
Usage
-----
Query Lambdas are named parameterized and versioned SQL queries
stored in Rockset and can be executed through a dedicated HTTPS endpoint.

Example
-------
::

    from rockset import Client, Q

    # connect securely to Rockset
    rs = Client()

    # create a Query Lambda
    qlambda = rs.QueryLambda.create(
        'myQueryLambda',
        query=Q('SELECT 1'))

    # print details about Query Lambda
    print(qlambda.workspace, qlambda.name, qlambda.version, qlambda.query)

    # execute a Query Lambda
    results = qlambda.execute()
    print(results)

.. _QueryLambda.create:

Create a new Query Lambda
-------------------------
Creating a Query Lambda using the Client_ object is as simple as
calling ``rs.QueryLambda.create('myQueryLambda', query=Q('SELECT...'))``::

    from rockset import Client, Q, F, P
    rs = Client()

    # construct a parameterized query
    q = Q('_events').where(F['type'] == P['target_type']).limit(100)

    # set default value for query parameter
    q.P['target_type'] = 'INFO'

    # create a new Query Lambda
    qlambda = rs.QueryLambda.create(
        'myQueryLambda',
        workspace='commons',
        query=q,
    )

.. _QueryLambda.list:

List all Query Lambdas
----------------------
List all Query Lambdas using the Client_ object using::

    from rockset import Client
    rs = Client()

    # List latest version of all Query Lambdas across all workspaces
    qlambdas = rs.QueryLambda.list()

    # List latest version of all Query Lambdas in a given workspace
    qlambdas = rs.QueryLambda.list(workspace='commons')

.. _QueryLambda.retrieveByTag:

Retrieve an existing Query Lambda version by a given tag
--------------------------------------------------------
Retrieve the version of a given Query Lambda associated with a given tag::

    from rockset import Client
    rs = Client()

    # retrieve Query Lambda
    qlambda = rs.QueryLambda.retrieveByTag(
        'myQueryLambda',
        tag='dev',
        workspace='commons')

.. _QueryLambda.retrieveByVersion:

Retrieve an existing Query Lambda version
-----------------------------------------
Retrieve a particular version of a given Query Lambda::

    from rockset import Client
    rs = Client()

    # retrieve Query Lambda
    qlambda = rs.QueryLambda.retrieveByVersion(
        'myQueryLambda',
        version='ac34bfg234ee',
        workspace='commons')

.. _QueryLambda.describe:

Describe an existing Query Lambda version
-----------------------------------------
Fetch all details about a particular version of a given Query Lambda::

    from rockset import Client
    rs = Client()

    # retrieve Query Lambda
    qlambda = rs.QueryLambda.retrieveByTag(
        'myQueryLambda',
        tag='dev',
        workspace='commons')

    # print all details about this Query Lambda version
    print(qlambda.describe())

.. _QueryLambda.execute:

Execute a specific Query Lambda version
--------------------------------------
Execute a Query Lambda version, optionally passing in parameters::

    from rockset import Client, ParamDict
    rs = Client()

    # retrieve Query Lambda
    qlambda = rs.QueryLambda.retrieveByTag(
        'myQueryLambda',
        tag='dev',
        workspace='commons')

    params = ParamDict()
    for target_type in ['INFO', 'DEBUG', 'WARN', 'ERROR']:
        params['target_type'] = target_type
        for result in qlambda.execute(parameters=params).results:
            print(target_type, result)

.. _QueryLambda.update:

Update a Query Lambda by creating a new version
-----------------------------------------------
Update the query associated with the Query Lambda by creating a
new version of it and optionally giving it a tag::

    from rockset import Client, Q, F, P, ParamDict
    rs = Client()

    # retrieve Query Lambda
    ql = rs.QueryLambda.retrieveByTag(
        'myQueryLambda',
        tag='dev',
        workspace='commons')

    # construct a new query
    q = (Q('_events')
        .where(
            (F['type'] == P['target_type']) &
            (F['userEmail'] == P['user_email'])
        )
        .limit(100)
    )

    # optionally, set default value for some or all of the query parameters
    q.P['target_type'] = 'INFO'

    # update Query Lambda
    # optionally, add a version tag at this time
    ql_new_version = ql.update(q, tag='dev')

    # execute the new Query Lambda for different user_emails
    params = ParamDict()
    for email in ['scott@rockset.com', 'veeve@rockset.com']:
        params['user_email'] = email
        results = ql_new_version.execute(params).results
        print(email, results)

.. _QueryLambda.tag:

Tag a version of a Query Lambda
-------------------------------
Apply a tag to a specific Query Lambda version::

    from rockset import Client
    rs = Client()

    # retrieve Query Lambda
    ql = rs.QueryLambda.retrieveByVersion(
        'myQueryLambda',
        version='ac34bfg234ee',
        workspace='commons')

    # add tag 'dev' to this Query Lambda version
    ql.tag('dev')

.. _QueryLambda.history:

List all versions of a Query Lambda
-----------------------------------
Fetch all versions of a given Query Lambda::

    from rockset import Client
    rs = Client()

    # retrieve Query Lambda
    ql = rs.QueryLambda.retrieveByTag(
        'myQueryLambda',
        tag='dev',
        workspace='commons')

    # fetch all versions of this Query Lambda
    all_qlambda_versions = ql.history()

.. _QueryLambda.drop:

Drop Query Lambda along with all previous versions
--------------------------------------------------
Use the ``drop()`` method to remove a Query Lambda permanently from Rockset.

.. note:: This is a permanent and non-recoverable operation. Beware.

::

    from rockset import Client
    rs = Client()

    qlambda = rs.QueryLambda.retrieveByVersion(
        'myQueryLambda',
        version='ac34bfg234ee',
        workspace='commons')
    qlambda.drop()

"""
from rockset.query import Q
from rockset.swagger_client.api import (QueryLambdasApi)
from rockset.swagger_client.models import (
    QueryLambda, QueryLambdaSql, CreateQueryLambdaRequest,
    CreateQueryLambdaTagRequest, UpdateQueryLambdaRequest,
    ExecuteQueryLambdaRequest, QueryParameter
)

import rockset


def _convert_query_to_qlsql(query):
    if not isinstance(query, rockset.Query):
        raise ValueError(
            "Input query of type `{}` is not supported. "
            "Please pass a valid rockset.Query object as input.".format(
                type(query)
            )
        )

    (sqltext, sqlparams) = query.sql()
    ql_params = []
    for param in sqlparams.sqlparams():
        ql_params.append(QueryParameter(**param))
    return QueryLambdaSql(query=sqltext, default_parameters=ql_params)


class QueryLambdaClient(object):
    def __init__(self, client):
        self._api = QueryLambdasApi(client)

    def create(
        self, name, query, workspace="commons", description=None, **kwargs
    ):
        # convert rockset.Query to QueryLambdaSql
        ql_sql = _convert_query_to_qlsql(query)

        # construct CreateQueryLambdaRequest
        request = CreateQueryLambdaRequest(
            name=name,
            description=description,
            sql=ql_sql,
        )

        # issue create api call
        ql_model = self._api.create(workspace=workspace,
                                    body=request).get('data')
        return QueryLambdaVersion(self._api, ql_model)

    def list(self, workspace=None):
        if workspace:
            # If workspace specified, list queries in that workspace
            ql_models = self._api.list_2(workspace=workspace).get('data')
        else:
            # Else list all queries across all workspaces in the org
            ql_models = self._api.list().get('data')
        return [QueryLambda(self._api, qlm) for qlm in ql_models]

    def retrieve(self, name, version=None, workspace='commons', **kwargs):
        if not version:
            raise ValueError('Query Lambda version not specified.')
        ql_model = self._api.get_0(
            workspace=workspace, query_lambda=name, version=version
        ).get('data')
        return QueryLambdaVersion(self._api, ql_model)

    def retrieveByVersion(
        self, name, version=None, workspace='commons', **kwargs
    ):
        return self.retrieve(name, version=version, workspace=workspace)

    def retrieveByTag(self, name, tag=None, workspace='commons', **kwargs):
        if not tag:
            raise ValueError('Query Lambda tag not specified.')
        ql_model = self._api.get(
            workspace=workspace, query_lambda=name, tag=tag
        ).get('data').get('version')
        return QueryLambdaVersion(self._api, ql_model)


class QueryLambda(rockset.swagger_client.models.QueryLambda):
    def __init__(self, api, model):
        super(QueryLambda, self).__init__(**model.to_dict())
        self._api = api
        self._model = model

    def history(self):
        """Returns the entire version history of this Query Lambda.

        Returns:
            list: List of QueryLambdaVersion objects each representing a particular
            version of this Query Lambda
        """
        ql_models = self._api.list_4(
            workspace=self.workspace, query_lambda=self.name
        ).get('data')
        return [QueryLambdaVersion(self._api, qlm) for qlm in ql_models]


class QueryLambdaVersion(rockset.swagger_client.models.QueryLambdaVersion):
    def __init__(self, api, model):
        super(QueryLambdaVersion, self).__init__(**model.to_dict())
        self._api = api
        self._model = model

        # pull out the required fields explicitly
        self._query = Q(model.sql.query)
        for param in model.sql.default_parameters:
            self._query.P[param.name] = param.value

    @property
    def query(self):
        """Returns the query as a rockset.Query object along with
        default parameters specified for this Query Lambda version.

        Returns:
            Query: a Query object that represents the backend query
            that will be run when this Query Lambda is executed.
        """
        return self._query

    def describe(self):
        """Returns all properties of the Query Lambda as a dict.

        Returns:
            dict: with all properties of the Query Lambda
        """
        return self.to_dict()

    def history(self):
        """Returns the entire version history of this Query Lambda.

        Returns:
            list: List of QueryLambdaVersion objects each representing a particular
            version of this Query Lambda
        """
        ql_models = self._api.list_4(
            workspace=self.workspace, query_lambda=self.name
        ).get('data')
        return [QueryLambdaVersion(self._api, qlm) for qlm in ql_models]

    def update(self, query, tag=None, description=None):
        """Updates the query associated with the Query Lambda and returns
        a new QueryLambda object that represents the newly created version of
        this Query Lambda. Optionally applies a tag to this new version.

        Args:
           query (Query): Query object using which a new version of this
           Query Lambda will be created.
           tag (str): Version tag that will be applied to the newly created version
           description (str): Human readable description that can be used as
           a commit message for this update.

        Returns:
            QueryLambdaVersion: a QueryLambdaVersion object that represents the new version
        """
        ql_sql = _convert_query_to_qlsql(query)
        request = UpdateQueryLambdaRequest(sql=ql_sql, description=description)
        ql_model = self._api.update(
            workspace=self.workspace, query_lambda=self.name, body=request
        ).get('data')
        if tag:
            tag_request = CreateQueryLambdaTagRequest(
                version=ql_model.version, tag_name=tag
            )
            self._api.create_0(
                workspace=self.workspace,
                query_lambda=self.name,
                body=tag_request
            )
        return QueryLambdaVersion(self._api, ql_model)

    def tag(self, tag):
        """Applies a specified tag to this Query Lambda version. If this tag is currently
        applied to another version of this Query Lambda, it will be moved. If not, it will
        be created.

        Args:
            tag (str): Version tag that wil be applied to this version.

        Returns:
            QueryLambdaVersion: the current QueryLambdaVersion object
        """
        tag_request = CreateQueryLambdaTagRequest(
            version=self.version, tag_name=tag
        )
        ql_model = self._api.create_0(
            workspace=self.workspace, query_lambda=self.name, body=tag_request
        ).get('data').get('version')
        return QueryLambdaVersion(self._api, ql_model)

    def execute(self, parameters=None, **kwargs):
        """Execute this Query Lambda using the given parameters and
        return a Cursor object to iterate through the results.

        If the Query Lambda was created without any parameters
        then no parameters need to be specified.

        If the Query Lambda was created with 1 or more query parameters,
        then all of those parameters can be specified as input.

        If a query parameter was not specified and a default value was
        provided for that query parameter when that Query Lambda version was
        created, then the default value will be used to execute the query.

        If a required query parameter was not specified,
        then an InputError Exception will be raised.

        Args:
            parameters (ParamDict): a ParamDict instance with values bound
            for all required query parameters, which will be used to
            execute this Query Lambda.

        Returns:
            Cursor: returns a cursor that can fetch query results
        """
        all_params = None
        if parameters:
            if not isinstance(parameters, rockset.ParamDict):
                raise ValueError(
                    "Query parameters of type `{}` is not supported. "
                    "Please pass a valid rockset.ParamDict object as input.".
                    format(type(parameters))
                )
            all_params = []
            for param in parameters.sqlparams():
                all_params.append(QueryParameter(**param))
        request = ExecuteQueryLambdaRequest(parameters=all_params, **kwargs)
        # TODO TODO TODO -- turn this into a Cursor
        return self._api.execute_0(
            workspace=self.workspace,
            query_lambda=self.name,
            version=self.version,
            body=request,
        )

    def drop(self):
        """Deletes the Query Lambda along with the current and all previous
        versions. Beware, this will permanently delete the Query Lambda and
        will be unrecoverable.
        """
        return self._api.delete(self.workspace, self.name).get('data')



"""DB-API implementation backed by Rockset

See http://www.python.org/dev/peps/pep-0249/

TODO:
- Handle DATE, TIME, INTERVAL (etc) data types

"""
from rockset import Client
from rockset.query import QueryStringSQLText
from rockset.sql.exception import *

import collections
import datetime
import rockset
import time

#
# Global constants
#
# supported DBAPI level
apilevel = '2.0'
# Python extended format codes, e.g. ...WHERE name=%(name)s
paramstyle = 'named'
# Threads may share the module and connections.
threadsafety = 2


def connect(*args, **kwargs):
    """Constructor for creating a connection to Rockset. Any argument that you
    pass to create a Rockset Client_ can be passed here.

    :returns: a :py:class:`Connection` object.
    """
    return Connection(*args, **kwargs)


class Connection(object):
    """Setup a Rockset Client handle
    """
    def __init__(self, *args, **kwargs):
        try:
            self._rs_client = Client(*args, **kwargs)
            self._rs_client.list()
        except rockset.exception.Error as e:
            raise Error.from_rockset_exception(e)

    def _client(self):
        """Return internal Rockset Client object"""
        return self._rs_client

    def close(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        raise NotSupportedError("Rockset does not support transactions")

    def cursor(self, cursor_type=None):
        if cursor_type is None:
            cursor_type = Cursor
        elif not issubclass(cursor_type, Cursor):
            raise ValueError(
                '{} is not a valid cursor type'.format(cursor_type)
            )
        return cursor_type(conn=self, client=self._rs_client)


class Cursor(object):
    """These objects represent a database cursor, which is used to manage the
    context of a fetch operation. Cursors created from the same connection are
    not isolated, i.e., any changes done to the database by a cursor are
    immediately visible by the other cursors.
    """
    def __init__(self, conn, client):
        """DBAPI Cursor
        """
        self._conn = conn
        self._rs_client = client
        self._arraysize = 1
        self._reset()

    def _reset(self):
        """Reset state of the Cursor back to empty"""
        self._cursor = None
        self._cursor_iter = None
        self._schema = {}
        self._fields = []
        self._warnings = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _get_fields(self):
        # nothing to do if no queries have been executed yet
        if self._cursor is None:
            return None

        return [(f['name'], f['type']) for f in self._cursor.fields()]

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result
        column:
            name
            type_code (Rockset data type name
                       see https://docs.rockset.com/data-types/)
            display_size (optional, always None for now)
            internal_size (optional, always None for now)
            precision (optional, always None for now)
            scale (optional, always None for now)
            null_ok (optional, always True for now)

        The first two items (name and type_code) are mandatory, the other
        five are optional and are set to None if no meaningful values can be
        provided.

        This attribute will be None for operations that do not return rows or
        if the cursor has not had an operation invoked via the .execute*()
        method yet.

        The type_code can be interpreted by comparing it to the Type Objects
        specified in the section below.
        """
        if self._cursor is None:
            return None

        return [
            # name, type_code, display_size, internal_size, precision, scale,
            # null_ok
            (f, t, None, None, None, None, (f != '_id'))
            for (f, t) in self._get_fields()
        ]

    @property
    def rowcount(self):
        if self._cursor is None:
            return -1
        return self._cursor.rowcount()

    def close(self):
        """Close the cursor now (rather than whenever __del__ is called).

        The cursor will be unusable from this point forward; an Error (or
        subclass) exception will be raised if any operation is attempted with
        the cursor.
        """
        self._reset()

    def execute(self, operation, parameters=None, generate_warnings=False):
        """Prepare and execute a database operation (query or command).
        Parameters may be provided as sequence or mapping and will be bound to
        variables in the operation. Variables are specified in ``pyformat``
        notation.

        Return values are not defined.
        """
        q = QueryStringSQLText(operation)
        if parameters:
            if not isinstance(parameters, dict):
                raise ProgrammingError(
                    "Unsupported type for query parameters: "
                    "expected: `dict` found: `{}`".format(type(parameters))
                )
            q.P.update(parameters)

        # setup the cursor
        self._cursor = self._rs_client.sql(
            q=q, generate_warnings=generate_warnings
        )
        # setup an iterator to fetch results
        self._cursor_iter = iter(self._cursor.results())
        self._warnings = self._cursor.warnings()

    def warnings(self):
        """Return warnings, if any, from the current query.
        """
        return self._warnings

    def executemany(operation, seq_of_parameters):
        """Prepare a database operation (query or command) and then execute it
        against all parameter sequences or mappings found in the sequence
        ``seq_of_parameters``.

        Only the final result set is retained.

        Return values are not defined.
        """
        for parameters in seq_of_parameters:
            self.execute(operation, parameters)

    def _fetchonedoc(self):
        if self._cursor is None:
            raise ProgrammingError(message='no query has been executed yet')
        try:
            one = next(self._cursor_iter)
        except StopIteration:
            return None
        except rockset.exception.Error as e:
            raise Error.from_rockset_exception(e)
        return one

    def fetchone(self):
        """Fetch the next row of a query result set, returning a single
        sequence, or None when no more data is available.

        An Error (or subclass) exception is raised if the previous call to
        .execute*() did not produce any result set or no call was issued yet.
        """
        one = self._fetchonedoc()
        if one is None:
            return None
        r = []
        for (f, t) in self._get_fields():
            if f in one:
                r.append(one[f])
            else:
                r.append(None)
        return tuple(r)

    def fetchmany(self, size=None):
        """Fetch the next set of rows of a query result, returning a sequence
        of sequences (e.g. a list of tuples). An empty sequence is returned
        when no more rows are available.

        The number of rows to fetch per call is specified by the parameter. If
        it is not given, the cursor's arraysize determines the number of rows
        to be fetched. The method should try to fetch as many rows as indicated
        by the size parameter. If this is not possible due to the
        specified number of rows not being available, fewer rows may be
        returned.

        An Error (or subclass) exception is raised if the previous call to
        .execute*() did not produce any result set or no call was issued yet.
        """
        if size is None:
            size = self.arraysize
        result = []
        for _ in range(size):
            one = self.fetchone()
            if one is None:
                break
            else:
                result.append(one)
        return result

    def fetchall(self):
        """Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples).

        An Error (or subclass) exception is raised if the previous call to
        .execute*() did not produce any result set or no call was issued yet.
        """
        result = []
        while True:
            one = self.fetchone()
            if one is None:
                break
            else:
                result.append(one)
        return result

    @property
    def arraysize(self):
        """This read/write attribute specifies the number of rows to fetch at a
        time with .fetchmany(). It defaults to 1 meaning to fetch a single row
        at a time.
        """
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        self._arraysize = value

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass

    # Optional DB-API Extensions
    @property
    def connection(self):
        return self._conn

    def __next__(self):
        """Return the next row from the currently executing SQL statement using
        the same semantics as .fetchone(). A ``StopIteration`` exception is
        raised when the result set is exhausted.
        """
        one = self.fetchone()
        if one is None:
            raise StopIteration
        else:
            return one

    next = __next__

    def __iter__(self):
        """Return self to make cursors compatible to the iteration protocol."""
        return self


class DictCursor(Cursor):
    def fetchone(self):
        return self._fetchonedoc()


#
# Type Objects
# See https://www.python.org/dev/peps/pep-0249/#implementation-hints-for-module-authors
#
class _DBAPITypeObject:
    def __init__(self, *values):
        self.values = values

    def __cmp__(self, other):
        if other in self.values:
            return 0
        if other < self.values:
            return 1
        else:
            return -1


Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime


def DateFromTicks(ticks):
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return Timestamp(*time.localtime(ticks)[:6])


Binary = str

STRING = _DBAPITypeObject([rockset.document.DATATYPE_STRING])
BINARY = _DBAPITypeObject([rockset.document.DATATYPE_BYTES])
NUMBER = _DBAPITypeObject(
    [rockset.document.DATATYPE_INT, rockset.document.DATATYPE_FLOAT]
)
DATETIME = _DBAPITypeObject(
    [
        rockset.document.DATATYPE_DATE, rockset.document.DATATYPE_DATETIME,
        rockset.document.DATATYPE_TIME, rockset.document.DATATYPE_TIMESTAMP
    ]
)
ROWID = _DBAPITypeObject()



"""
Introduction
------------
Various Python exceptions thrown by the ``rockset`` module are explained in
this section, along with possible reasons and remedies to assist in
trouble-shooting.

.. _AuthError:

Authentication Errors
---------------------
The server is rejecting your request because you have either an expired or
an invalid token. Ensure you have a valid API key or generate a new one
using the Rockset Console before trying your request again.

.. autoclass:: rockset.exception.AuthError

.. _InputError:

Input Errors
------------
The server is unable to understand the API request as it was sent. This most
likely means the API was badly formed (like the input query has a syntax error).
When you encounter this error, please refer to the relevant documentation and
verify if the request is constructed properly and if the resource is still
present.

.. autoclass:: rockset.exception.InputError

.. _LimitReached:

Limit Reached
-------------
The server could understand the input request but refuses to execute it.
This commonly happens when an account limit has been reached. Please reach
out to Rockset Support with more details to alter your account limit.

.. autoclass:: rockset.exception.LimitReached

.. _NotYetImplemented:

Not Yet Implemented
-------------------
Your API request needs a feature that is not present in your cluster for it to
complete. Your cluster needs an upgrade or this feature is in our roadmap but
we haven't gotten around to implementing it yet. Please reach out to Rockset
support with more details to help us prioritize this feature.

.. autoclass:: rockset.exception.NotYetImplemented

.. _RequestTimeout:

Request Timeouts
----------------
The server did not complete the API request before the timeout you set for
the request expired. To troubleshoot, see if your request succeeds when you
don't set a timeout. If it does then you need to recalibrate your timeout value.
If it doesn't, then debug the issue based on the new error you receive.

.. autoclass:: rockset.exception.RequestTimeout

.. _ServerError:

Server Errors
-------------
These errors mean the server correctly parsed the input request, but couldn't
process it for some reason.  If a particular request or application is seeing
this while other requests are fine, then you probably uncovered a bug with
Rockset. Please contact Rockset support to report the bug and we will provide
a time estimte for resolution and send you a t-shirt.

.. autoclass:: rockset.exception.ServerError

.. _TransientServerError:

Transient Server Errors
-----------------------

When many of your requests are failing with TransientServerErrors, it means our
servers are going through a period of instability or unplanned downtime.
This always means our alerts are firing, our pagers are ringing, phones are
buzzing, and little adorable kittens are getting lost in the woods. We are
actively investigating and fixing this issue. Look for upates on our status
page with estimates on time to resolution. Sorry.

.. autoclass:: rockset.exception.TransientServerError

"""
import json


class Error(Exception):
    """Base class for all rockset exceptions"""
    def __init__(self, **kwargs):
        message = kwargs.pop('message', None)
        if message is None:
            message = 'Unexpected error'
        super(Error, self).__init__(message)
        kwargs['message'] = message
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        parts = []
        if hasattr(self, 'code'):
            parts.append('{}'.format(self.code))
        if hasattr(self, 'type'):
            parts.append('{}'.format(self.type))
        parts.append(self.message)
        return ' '.join(parts)


class HTTPError(Error):
    """Error returned by swagger API call.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    def __init__(self, **kwargs):
        kwargs['message'] = kwargs.get(
            'message', 'Unexpected error from server'
        )
        super(HTTPError, self).__init__(**kwargs)


class AuthError(HTTPError):
    """API key or access token is missing, expired or invalid.
    Re-authenticating with a valid API key should normally fix this.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
    """
    pass


class InputError(HTTPError):
    """User request has a missing or invalid parameter and cannot be
    processed as is. Syntax errors in queries fall in this category.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    pass


class LimitReached(HTTPError):
    """The API request has exceeded some user-defined limit (such as
    max deadline set for a query) or a system limit. Refer to
    documentation to increase the limit or reach out to Rockset support
    with more details to alter your account limit.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    pass


class NotYetImplemented(HTTPError):
    """Your request is expecting a feature that has not been deployed in your
    cluster or has not yet been implemented. Please reach out to Rockset
    support with more details to help us prioritize this feature. Thank you.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    pass


class ServerError(HTTPError):
    """Something totally unexpected happened on our servers while processing
    your request and most likely you have encountered a bug in Rockset.
    Please contact Rockset support and provide all the details you received
    along with the error for quick diagnosis, resolution, and to collect your
    t-shirt.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    pass


class TransientServerError(HTTPError):
    """Some transient hiccup made us fail this request. This means our
    oncall engineers are actively working on this issue and should resolve
    the issue soon. Please retry after sometime. Sorry.

    Attributes:
        code (int): HTTP status code obtained from server
        message (str): error message with more details
        type (str): error sub-category
    """
    pass


class ResourceSuspendedError(HTTPError):
    """ Target collection was suspended due to inactivity. A query
    issued against the resource will automatically trigger reactivation of
    the target collection. Please retry the query once the resource
    has been unsuspended. Use the Rockset Console or the CLI or contact
    Rockset support to configure the auto suspend behavior.

    Attributes:
        message (str): error message
    """
    pass


class RequestTimeout(Error):
    """Request timed out.

    Many API calls allow a client side timeout to be specified.
    When specified, this exception will be thrown when the timeout
    expires and the API call has not received a valid response
    or an error from the servers.

    Attributes:
        message (str): timeout error message
        timeout (int): timeout specfied with API call in seconds
    """
    def __init__(self, **kwargs):
        kwargs['message'] = kwargs.get('message', 'Request timed out')
        kwargs['timeout'] = kwargs.get('timeout', None)
        super(RequestTimeout, self).__init__(**kwargs)


__all__ = [
    'AuthError',
    'InputError',
    'LimitReached',
    'NotYetImplemented',
    'RequestTimeout',
    'ServerError',
    'TransientServerError',
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rockset/document.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rockset import value_pb2 as rockset_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rockset/document.proto',
  package='rockset',
  syntax='proto3',
  serialized_options=b'\n\020io.rockset.protoP\000\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16rockset/document.proto\x12\x07rockset\x1a\x13rockset/value.proto\"I\n\x08\x44ocument\x12\n\n\x02id\x18\x02 \x01(\t\x12\x12\n\nevent_time\x18\x01 \x01(\x03\x12\x1d\n\x05value\x18\x03 \x01(\x0b\x32\x0e.rockset.ValueB\x17\n\x10io.rockset.protoP\x00\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[rockset_dot_value__pb2.DESCRIPTOR,])




_DOCUMENT = _descriptor.Descriptor(
  name='Document',
  full_name='rockset.Document',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='rockset.Document.id', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_time', full_name='rockset.Document.event_time', index=1,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='rockset.Document.value', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=56,
  serialized_end=129,
)

_DOCUMENT.fields_by_name['value'].message_type = rockset_dot_value__pb2._VALUE
DESCRIPTOR.message_types_by_name['Document'] = _DOCUMENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Document = _reflection.GeneratedProtocolMessageType('Document', (_message.Message,), {
  'DESCRIPTOR' : _DOCUMENT,
  '__module__' : 'rockset.document_pb2'
  # @@protoc_insertion_point(class_scope:rockset.Document)
  })
_sym_db.RegisterMessage(Document)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import unittest

def my_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = unittest.TestSuite()
    test_suite.addTest(test_loader.discover('rockset.tests'))
    test_suite.addTest(test_loader.discover('rockset.sql.tests'))
    test_suite.addTest(test_loader.discover('rockset.rock.tests'))
    return test_suite



"""
Package private common utilities. Do not use directly.
"""
from __future__ import absolute_import
from __future__ import unicode_literals
import rockset

__all__ = [
    'Error',
    'Warning',
    'InterfaceError',
    'DatabaseError',
    'DataError',
    'OperationalError',
    'IntegrityError',
    'InternalError',
    'ProgrammingError',
    'NotSupportedError',
]


class Warning(Exception):
    """Exception raised for important warnings like data truncations while inserting, etc."""
    pass


class Error(rockset.exception.Error):
    """Exception that is the base class of all other error exceptions. You
    can use this to catch all errors with one single except statement.
    Warnings are not considered errors and thus should not use this class as
    base.
    """
    @classmethod
    def from_rockset_exception(cls, rse, **kwargs):
        kwargs['message'] = rse.message
        kwargs['code'] = hasattr(rse, 'code') and rse.code or None
        kwargs['type'] = hasattr(rse, 'type') and rse.type or None
        if type(rse) == rockset.exception.ServerError:
            dbapi_exception = InternalError(**kwargs)
        elif type(rse) == rockset.exception.NotYetImplemented:
            dbapi_exception = NotSupportedError(**kwargs)
        elif type(rse) == rockset.exception.AuthError:
            dbapi_exception = OperationalError(**kwargs)
        elif type(rse) == rockset.exception.LimitReached:
            dbapi_exception = OperationalError(**kwargs)
        elif type(rse) == rockset.exception.ResourceSuspendedError:
            dbapi_exception = OperationalError(**kwargs)
        elif type(rse) == rockset.exception.RequestTimeout:
            dbapi_exception = OperationalError(**kwargs)
        elif type(rse) == rockset.exception.TransientServerError:
            dbapi_exception = OperationalError(**kwargs)
        elif type(rse) == rockset.exception.InputError:
            dbapi_exception = ProgrammingError(**kwargs)
        else:
            dbapi_exception = cls(**kwargs)
        return dbapi_exception


class InterfaceError(Error):
    """Exception raised for errors that are related to the database interface
    rather than the database itself.
    """
    pass


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""
    pass


class DataError(DatabaseError):
    """Exception raised for errors that are due to problems with the processed
    data like division by zero, numeric value out of range, etc.
    """
    pass


class IntegrityError(DatabaseError):
    """Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc
    """
    pass


class InternalError(DatabaseError):
    """Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc
    """
    pass


class NotSupportedError(DatabaseError):
    """Exception raised in case a method or database API was used which is not
    supported by the database, e.g. requesting a ``.rollback()`` on a
    connection that does not support transaction or has transactions turned off
    """
    pass


class OperationalError(DatabaseError):
    """Exception raised for errors that are related to the database's
    operation and not necessarily under the control of the programmer,
    e.g. an unexpected disconnect occurs, the data source name is not found,
    a transaction could not be processed, a memory allocation error occurred
    during processing, etc.
    """
    pass


class ProgrammingError(DatabaseError):
    """Exception raised for programming errors, e.g. table not found or
    already exists, syntax error in the SQL statement, wrong number of
    parameters specified, etc.
    """
    pass



# coding: utf-8

"""
    REST API

    Rockset's REST API allows for creating and managing all resources in Rockset. Each supported endpoint is documented below.  All requests must be authorized with a Rockset API key, which can be created in the [Rockset console](https://console.rockset.com). The API key must be provided as `ApiKey <api_key>` in the `Authorization` request header. For example: ``` Authorization: ApiKey aB35kDjg93J5nsf4GjwMeErAVd832F7ad4vhsW1S02kfZiab42sTsfW5Sxt25asT ```  All endpoints are only accessible via https.  Build something awesome!  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class CsvParams(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'first_line_as_column_names': 'bool',
        'separator': 'str',
        'encoding': 'str',
        'column_names': 'list[str]',
        'column_types': 'list[str]',
        'quote_char': 'str',
        'escape_char': 'str'
    }

    attribute_map = {
        'first_line_as_column_names': 'firstLineAsColumnNames',
        'separator': 'separator',
        'encoding': 'encoding',
        'column_names': 'columnNames',
        'column_types': 'columnTypes',
        'quote_char': 'quoteChar',
        'escape_char': 'escapeChar'
    }

    def __init__(self, **kwargs):  # noqa: E501
        """CsvParams - a model defined in Swagger"""  # noqa: E501

        self._first_line_as_column_names = None
        self._separator = None
        self._encoding = None
        self._column_names = None
        self._column_types = None
        self._quote_char = None
        self._escape_char = None
        self.discriminator = None

        self.first_line_as_column_names = kwargs.pop('first_line_as_column_names', None)
        self.separator = kwargs.pop('separator', None)
        self.encoding = kwargs.pop('encoding', None)
        self.column_names = kwargs.pop('column_names', None)
        self.column_types = kwargs.pop('column_types', None)
        self.quote_char = kwargs.pop('quote_char', None)
        self.escape_char = kwargs.pop('escape_char', None)

    @property
    def first_line_as_column_names(self):
        """Gets the first_line_as_column_names of this CsvParams.  # noqa: E501

        If the first line in every object specifies the column names  # noqa: E501

        :return: The first_line_as_column_names of this CsvParams.  # noqa: E501
        :rtype: bool
        """
        return self._first_line_as_column_names

    @first_line_as_column_names.setter
    def first_line_as_column_names(self, first_line_as_column_names):
        """Sets the first_line_as_column_names of this CsvParams.

        If the first line in every object specifies the column names  # noqa: E501

        :param first_line_as_column_names: The first_line_as_column_names of this CsvParams.  # noqa: E501
        :type: bool
        """

        self._first_line_as_column_names = first_line_as_column_names

    @property
    def separator(self):
        """Gets the separator of this CsvParams.  # noqa: E501

        a single character that is the column separator  # noqa: E501

        :return: The separator of this CsvParams.  # noqa: E501
        :rtype: str
        """
        return self._separator

    @separator.setter
    def separator(self, separator):
        """Sets the separator of this CsvParams.

        a single character that is the column separator  # noqa: E501

        :param separator: The separator of this CsvParams.  # noqa: E501
        :type: str
        """

        self._separator = separator

    @property
    def encoding(self):
        """Gets the encoding of this CsvParams.  # noqa: E501

        can be one of: UTF-8, ISO_8859_1, UTF-16  # noqa: E501

        :return: The encoding of this CsvParams.  # noqa: E501
        :rtype: str
        """
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        """Sets the encoding of this CsvParams.

        can be one of: UTF-8, ISO_8859_1, UTF-16  # noqa: E501

        :param encoding: The encoding of this CsvParams.  # noqa: E501
        :type: str
        """

        self._encoding = encoding

    @property
    def column_names(self):
        """Gets the column_names of this CsvParams.  # noqa: E501

        names of columns  # noqa: E501

        :return: The column_names of this CsvParams.  # noqa: E501
        :rtype: list[str]
        """
        return self._column_names

    @column_names.setter
    def column_names(self, column_names):
        """Sets the column_names of this CsvParams.

        names of columns  # noqa: E501

        :param column_names: The column_names of this CsvParams.  # noqa: E501
        :type: list[str]
        """

        self._column_names = column_names

    @property
    def column_types(self):
        """Gets the column_types of this CsvParams.  # noqa: E501

        names of columns  # noqa: E501

        :return: The column_types of this CsvParams.  # noqa: E501
        :rtype: list[str]
        """
        return self._column_types

    @column_types.setter
    def column_types(self, column_types):
        """Sets the column_types of this CsvParams.

        names of columns  # noqa: E501

        :param column_types: The column_types of this CsvParams.  # noqa: E501
        :type: list[str]
        """

        self._column_types = column_types

    @property
    def quote_char(self):
        """Gets the quote_char of this CsvParams.  # noqa: E501

        character within which a cell value is enclosed,null character if no such character, default is '\"'  # noqa: E501

        :return: The quote_char of this CsvParams.  # noqa: E501
        :rtype: str
        """
        return self._quote_char

    @quote_char.setter
    def quote_char(self, quote_char):
        """Sets the quote_char of this CsvParams.

        character within which a cell value is enclosed,null character if no such character, default is '\"'  # noqa: E501

        :param quote_char: The quote_char of this CsvParams.  # noqa: E501
        :type: str
        """

        self._quote_char = quote_char

    @property
    def escape_char(self):
        """Gets the escape_char of this CsvParams.  # noqa: E501

        escape character removes any special meaning from the following character,default is '\\'  # noqa: E501

        :return: The escape_char of this CsvParams.  # noqa: E501
        :rtype: str
        """
        return self._escape_char

    @escape_char.setter
    def escape_char(self, escape_char):
        """Sets the escape_char of this CsvParams.

        escape character removes any special meaning from the following character,default is '\\'  # noqa: E501

        :param escape_char: The escape_char of this CsvParams.  # noqa: E501
        :type: str
        """

        self._escape_char = escape_char

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(CsvParams, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CsvParams):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, item):
        return getattr(self, item)

    def items(self):
        return self.to_dict().items()

    def __setitem__(self, item, value):
        return seattr(self, item, value)



"""
Introduction
------------
Cursor objects are return by the Client.sql() API.
A cursor simply binds the query to a particular collection and the
query is not executed server side until the application starts to fetch
results from the cursor.

.. note:: Cursors are never instantiated directly by the application \
and are always instantiated by Rockset Client ``sql()`` APIs.

Fetch all results
-----------------

Use the cursor's results() method to fetch all the results of the query
in one shot::

    results = mycollection.sql(q).results()

The above technique will work well, if the number of results returned by the
query is limited, say because it employs a LIMIT clause.

For queries, that return a large number of results, please use the cursor
iterators as described below.

Iterators with automatic pagination
-----------------------------------

Cursor objects are iterables, so you can do something like::

    results = mycollection.sql(q)
    for r in results:
        print(r)

Cursor objects support seamless automatic pagination to iterate over large
result sets. The cursor iterator will fetch and buffer a small portion of the
results and as the iterator reaches the end of the current batch of buffered
results, it will automatically issue the query with the appropriate pagination
parameters to buffer the next batch and seamlessly continue results iteration.

The default cursor iterator uses a batch size of 10,000. You can create a cursor
iterator with a different batch size by using the ``iter()`` method.

Example using the default cursor iterator::

    results = mycollection.sql(q)
    for r in results:
        print(r)


Example using a custom cursor iterator with batch size 200::

    results = mycollection.sql(q)
    for r in results.iter(200):
        print(r)

.. automethod:: rockset.Cursor.iter()

Async requests
--------------

Cursors support asyncio.Future to schedule and run queries concurrently
along with other async events in your application.

One can create an asyncio.Future from a cursor object using the
``Cursor.async_request()`` method. These futures are not scheduled in
any async threads and the application have to schedule them in an asyncio
event loop. Once the futures are scheduled and run to completion, then
the results of their respective queries can be accessed by calling
future.result(). The return value of future.result() will be identical
to calling Cursor.results() API on the original query.

For example::

    jims_future = people.sql(F["first_name"] == "Jim").async_request()
    sfmerch_future = merchants.sql(F["zipcode"] == "94107").async_request()

    # can schedule these asyncio.futures along with other futures
    # issue both queries concurrently and block until both of them finish
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(jims_future, sfmerch_future))

    all_jims = jims_future.result()
    all_sfmerchants = sfmerch_future.result()

Even if an future was originally issued by an async_requst() API call,
one can still call the blocking Cursor.results() API to fetch the results
synchronously. Cursor.results() will schedule the future, block until the
future finishes execution and then will return those results. A subsequent
future.result() call will return the query results immediately.

For example::

    jims_cursor = people.sql(F["first_name"] == "Jim")
    jims_future = jims_cursor.async_request()

    # do a blocking results() will block on the future behind the scenes
    results = jims_cursor.results()

    # this will return immediately without incurring any server round-trip
    results2 = jims_future.result()


.. automethod:: rockset.Cursor.async_request()

"""

from rockset.document import Document
from rockset.exception import ServerError, InputError
from rockset.query import LimitQuery, SubQuery
from rockset.swagger_client.api.queries_api import QueriesApi
from rockset.swagger_client.models.query_request import QueryRequest
from rockset.swagger_client.models.query_response import QueryResponse

import asyncio
import sys


class Cursor(object):
    """ Fetch the results of a query executed against a collection
    """
    def __init__(
        self, q=None, client=None, generate_warnings=False, timeout=None
    ):
        self._query = q
        self._client = client
        self._lastq_response = None
        self._fullq_response = None
        self._future = None
        self._warnings = None
        self._generate_warnings = generate_warnings
        self._timeout = timeout

    def _run(self):
        if self._fullq_response is None:
            self._fullq_response = self._run_query()
        return self._fullq_response

    def _execute_query(self, request):
        """Keep all backend interaction within this function,
        so that it is easy to mock in tests"""
        # execute query
        qreq = QueryRequest(sql=request)
        response = QueriesApi(self._client
                             ).query(qreq, _request_timeout=self._timeout)
        if not isinstance(response, QueryResponse):
            raise ServerError(
                message='invalid return message (type={} results={})'.
                format(type(response), response)
            )
        return response

    def _run_query(self, limit=None, skip=0):
        # add limit, offset to query if required
        query = self._query
        if limit is not None:
            query = LimitQuery(
                limit=limit, skip=skip, child=SubQuery(query, alias='subq')
            )

        # build the query request object
        request = {}
        (sqltxt, params) = query.sql()
        request = {
            'query': sqltxt,
            'parameters': params.sqlparams(),
        }

        # Only add this parameters if `_generate_warnings` is set to True.
        # This is to ensure that if apiserver is not pushed, not requesting
        # warnings still works.
        # TODO(hieup): Put `generate_warnings` to `qo` once we push apiserver
        if self._generate_warnings:
            request['generate_warnings'] = True

        self._lastq_response = self._execute_query(request)
        return self._lastq_response

    def _get_docs(self, response):
        if not response.results:
            return []
        return [Document(row) for row in response.results]

    def __iter__(self):
        return self.iter()

    def __str__(self):
        return str(self.results())

    def __getitem__(self, key):
        return self.results()[key]

    def iter(self, batch=10000):
        """ Returns an iterator that does seamless automatic pagination
        behind the scenes while fetching no more than the specified
        batch size number of results at a time.

        Args:
            batch (int): maximum number of results fetched at a time

        Returns:
            Iterator Object: Iterator that will return all results one \
            :class:`Document` object at a time with automatic pagination
        """
        if batch <= 0:
            raise InputError(
                message='invalid cursor iterator batchsize. "{}" is not '
                'a positive integer.'.format(batch)
            )

        skip = 0
        while True:
            results = self._get_docs(self._run_query(batch, skip))
            for r in results:
                yield r
            if ((batch is None) or (batch > len(results))):
                break
            skip += batch

    def results(self):
        """ Execute the query and fetch all the results in one shot.

        Returns:
            list[:class:`Document`]: Results of the query represent as a list
            of Document objects
        """
        return self._get_docs(self._run())

    def warnings(self):
        """ Retrieve warnings that occurred during query execution, if any.

        Returns:
            list[str]: The warnings from the query. Returns None if no warnings exist.
        """
        return self._lastq_response and self._lastq_response.warnings

    def rowcount(self):
        """Number of rows returned by the last query

        Returns:
            int: The number of rows that the last query produced.
            Returns None in case no query has been executed yet.
        """
        return self._lastq_response and len(self._lastq_response.results)

    def fields(self):
        """ Once query has been executed, this method will return all the
        fields that are present in the query result, while preserving
        the order in which those fields were selected in the input query.

        NOTE: this method will return None when called before the query has
              actually been executed.

        Returns:
            list[dicts]: List of column headers in the query result,
                            preserving the ordering from the input query.
        """
        if not self._lastq_response:
            raise InputError(
                message='cannot call fields() before query execution'
            )

        # server will return field names in the order the query requested
        # for all queries except 'SELECT * FROM ...'
        columns = None
        if self._lastq_response.column_fields:
            columns = [cf['name'] for cf in self._lastq_response.column_fields]

        # TODO: come up with a better way to do this
        # get all results and infer schema from them
        # build a row with all possible fields, by merging all into one
        # remove this once apiserver returns both field names and types
        schema = Document()
        for r in self._get_docs(self._run()):
            schema.update(r)

        return schema.fields(columns=columns)

    def stats(self):
        """Query execution stats

        Returns:
            dict: Some key execution stats
            Returns None in case no query has been executed yet.
        """
        return self._lastq_response and self._lastq_response.stats

    def async_request(self):
        """ Returns an asyncio.Future object that can be scheduled in an
        asyncio event loop. Once scheduled and run to completion, the results
        can be fetched via the future.result() API. The return value of
        future.result() will be the same as the return value of Cursor.results()

        Returns:
            asyncio.Future: Returns a Future object that can be scheduled
            in an asyncio event loop and future.result() will hold the same
            return value as Cursor.results()
        """
        if self._future is not None:
            return self._future
        self._future = asyncio.Future()
        asyncio.ensure_future(self._issue())
        return self._future

    async def _issue(self):
        self._future.set_result(self.results())


__all__ = [
    'Cursor',
]



"""
Introduction
------------
Document object represents a single record or row in the result retrieved
from executing a query.

.. note:: Document objects are generally instantiated by the Cursor object \
and do not need to be instantiated directly by the application while \
retrieving results of a query.

This class adapts Rockset's SQL types to standard Python types for all the
top level fields retrieved by the query.

+----------------------+--------------------+-------------------------------------+
| Rockset Data Type    | Python Data Type   | Comments                            |
+======================+====================+=====================================+
| SQL NULL             | None               |                                     |
+----------------------+--------------------+-------------------------------------+
| JSON NULL            | None               |                                     |
+----------------------+--------------------+-------------------------------------+
| bool                 | bool               |                                     |
+----------------------+--------------------+-------------------------------------+
| int                  | int                |                                     |
+----------------------+--------------------+-------------------------------------+
| float                | float              |                                     |
+----------------------+--------------------+-------------------------------------+
| string               | str                |                                     |
+----------------------+--------------------+-------------------------------------+
| bytes                | str                | - NOTE: will be changed to bytes    |
|                      |                    |   in the future                     |
+----------------------+--------------------+-------------------------------------+
| array                | list               |                                     |
+----------------------+--------------------+-------------------------------------+
| object               | dict               |                                     |
+----------------------+--------------------+-------------------------------------+
| date                 | datetime.date      |                                     |
+----------------------+--------------------+-------------------------------------+
| datetime             | datetime.datetime  | tzinfo will be None                 |
+----------------------+--------------------+-------------------------------------+
| time                 | datetime.time      |                                     |
+----------------------+--------------------+-------------------------------------+
| timestamp            | str                | - ISO8601 format in UTC timezone    |
|                      |                    | - eg: '2019-11-09T23:14:31.561512Z' |
|                      |                    | - NOTE: will be changed to          |
|                      |                    |   datetime.datetime with tzinfo     |
|                      |                    |   in the future                     |
+----------------------+--------------------+-------------------------------------+
| month_interval       | dict               | - INTERVAL 10 MONTH                 |
|                      |                    |   will return {'value': 10}         |
|                      |                    | - INTERVAL '3-4' YEAR TO MONTH      |
|                      |                    |   will return {'value': 40}         |
+----------------------+--------------------+-------------------------------------+
| microsecond_interval | datetime.timedelta |                                     |
+----------------------+--------------------+-------------------------------------+
| geography.Point      | geojson.Point      |                                     |
+----------------------+--------------------+-------------------------------------+
| geography.LineString | geojson.LineString |                                     |
+----------------------+--------------------+-------------------------------------+
| geography.Polygon    | geojson.Polygon    |                                     |
+----------------------+--------------------+-------------------------------------+


.. note:: Please note that this type adaptation is only done for the top \
level fields returned in a query. If a top level field retrieved by the query \
is a map or an array, then fields nested within that map or an array are \
not adapted to the respective Python data types.

"""

import datetime
import geojson

# all Rockset data types
DATATYPE_META = '__rockset_type'
DATATYPE_INT = 'int'
DATATYPE_FLOAT = 'float'
DATATYPE_BOOL = 'bool'
DATATYPE_STRING = 'string'
DATATYPE_BYTES = 'bytes'
DATATYPE_NULL = 'null'
DATATYPE_NULL_TYPE = 'null_type'
DATATYPE_ARRAY = 'array'
DATATYPE_OBJECT = 'object'
DATATYPE_DATE = 'date'
DATATYPE_DATETIME = 'datetime'
DATATYPE_TIME = 'time'
DATATYPE_TIMESTAMP = 'timestamp'
DATATYPE_MONTH_INTERVAL = 'month_interval'
DATATYPE_MICROSECOND_INTERVAL = 'microsecond_interval'
DATATYPE_GEOGRAPHY = 'geography'


def _date_fromisoformat(s):
    dt = datetime.datetime.strptime(s, '%Y-%m-%d')
    return datetime.date(year=dt.year, month=dt.month, day=dt.day)


def _time_fromisoformat(s):
    try:
        dt = datetime.datetime.strptime(s, '%H:%M:%S.%f')
    except ValueError:
        dt = datetime.datetime.strptime(s, '%H:%M:%S')
    return datetime.time(
        hour=dt.hour,
        minute=dt.minute,
        second=dt.second,
        microsecond=dt.microsecond
    )


def _datetime_fromisoformat(s):
    try:
        dt = datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        dt = datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S')
    return dt


def _timedelta_from_microseconds(us):
    return datetime.timedelta(microseconds=us)


class Document(dict):
    """Represents a single record or row in query results.

    This is a sub-class of dict. So, treat this object as a dict
    for all practical purposes.

    Only the constructor is overridden to handle the type adaptations
    shown in the table above.
    """
    def __init__(self, *args, **kwargs):
        """This is a sub-class of dict. So, treat this object as a dict
        for all practical purposes.
        Only contains the constructor to handle the type adaptations
        shown in the table above.
        """
        super(Document, self).__init__(*args, **kwargs)
        for k in self.keys():
            if not isinstance(self[k], dict):
                continue
            if DATATYPE_META not in self[k]:
                continue
            t = self[k][DATATYPE_META].lower()
            v = self[k]['value']
            if t == DATATYPE_DATE:
                self[k] = _date_fromisoformat(v)
            elif t == DATATYPE_TIME:
                self[k] = _time_fromisoformat(v)
            elif t == DATATYPE_DATETIME:
                self[k] = _datetime_fromisoformat(v)
            elif t == DATATYPE_MICROSECOND_INTERVAL:
                self[k] = _timedelta_from_microseconds(v)
            elif t == DATATYPE_GEOGRAPHY:
                self[k] = geojson.GeoJSON.to_instance(v)

    def _py_type_to_rs_type(self, v):
        if isinstance(v, bool):
            # check for bool before int, since bools are ints too
            return DATATYPE_BOOL
        elif isinstance(v, int):
            return DATATYPE_INT
        elif isinstance(v, float):
            return DATATYPE_FLOAT
        elif isinstance(v, str):
            return DATATYPE_STRING
        elif isinstance(v, bytes):
            return DATATYPE_BYTES
        elif isinstance(v, type(None)):
            return DATATYPE_NULL
        elif isinstance(v, list):
            return DATATYPE_ARRAY
        elif isinstance(v, datetime.datetime):
            # check for datetime first, since datetimes are dates too
            return DATATYPE_DATETIME
        elif isinstance(v, datetime.date):
            return DATATYPE_DATE
        elif isinstance(v, datetime.time):
            return DATATYPE_TIME
        elif isinstance(v, datetime.timedelta):
            return DATATYPE_MICROSECOND_INTERVAL
        elif isinstance(v, geojson.GeoJSON):
            return DATATYPE_GEOGRAPHY
        elif isinstance(v, dict):  # keep this in the end
            if DATATYPE_META not in v:
                return DATATYPE_OBJECT
            return v[DATATYPE_META].lower()

    def fields(self, columns=None):
        columns = columns or sorted(self)
        return [
            {
                'name': c,
                'type': self._py_type_to_rs_type(self.get(c, type(None)))
            } for c in columns
        ]


__all__ = [
    'Document',
]



# coding: utf-8

"""
    REST API

    Rockset's REST API allows for creating and managing all resources in Rockset. Each supported endpoint is documented below.  All requests must be authorized with a Rockset API key, which can be created in the [Rockset console](https://console.rockset.com). The API key must be provided as `ApiKey <api_key>` in the `Authorization` request header. For example: ``` Authorization: ApiKey aB35kDjg93J5nsf4GjwMeErAVd832F7ad4vhsW1S02kfZiab42sTsfW5Sxt25asT ```  All endpoints are only accessible via https.  Build something awesome!  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from rockset.swagger_client.models.collection import Collection  # noqa: F401,E501
from rockset.swagger_client.models.dynamodb_integration import DynamodbIntegration  # noqa: F401,E501
from rockset.swagger_client.models.gcs_integration import GcsIntegration  # noqa: F401,E501
from rockset.swagger_client.models.kafka_integration import KafkaIntegration  # noqa: F401,E501
from rockset.swagger_client.models.kinesis_integration import KinesisIntegration  # noqa: F401,E501
from rockset.swagger_client.models.mongo_db_integration import MongoDbIntegration  # noqa: F401,E501
from rockset.swagger_client.models.redshift_integration import RedshiftIntegration  # noqa: F401,E501
from rockset.swagger_client.models.s3_integration import S3Integration  # noqa: F401,E501
from rockset.swagger_client.models.segment_integration import SegmentIntegration  # noqa: F401,E501


class Integration(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'description': 'str',
        'collections': 'list[Collection]',
        'created_by': 'str',
        'created_at': 'str',
        's3': 'S3Integration',
        'kinesis': 'KinesisIntegration',
        'dynamodb': 'DynamodbIntegration',
        'redshift': 'RedshiftIntegration',
        'gcs': 'GcsIntegration',
        'segment': 'SegmentIntegration',
        'kafka': 'KafkaIntegration',
        'mongodb': 'MongoDbIntegration'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'collections': 'collections',
        'created_by': 'created_by',
        'created_at': 'created_at',
        's3': 's3',
        'kinesis': 'kinesis',
        'dynamodb': 'dynamodb',
        'redshift': 'redshift',
        'gcs': 'gcs',
        'segment': 'segment',
        'kafka': 'kafka',
        'mongodb': 'mongodb'
    }

    def __init__(self, name, created_by, **kwargs):  # noqa: E501
        """Integration - a model defined in Swagger"""  # noqa: E501

        self._name = None
        self._description = None
        self._collections = None
        self._created_by = None
        self._created_at = None
        self._s3 = None
        self._kinesis = None
        self._dynamodb = None
        self._redshift = None
        self._gcs = None
        self._segment = None
        self._kafka = None
        self._mongodb = None
        self.discriminator = None

        self.name = name
        self.description = kwargs.pop('description', None)
        self.collections = kwargs.pop('collections', None)
        self.created_by = created_by
        self.created_at = kwargs.pop('created_at', None)
        self.s3 = kwargs.pop('s3', None)
        self.kinesis = kwargs.pop('kinesis', None)
        self.dynamodb = kwargs.pop('dynamodb', None)
        self.redshift = kwargs.pop('redshift', None)
        self.gcs = kwargs.pop('gcs', None)
        self.segment = kwargs.pop('segment', None)
        self.kafka = kwargs.pop('kafka', None)
        self.mongodb = kwargs.pop('mongodb', None)

    @property
    def name(self):
        """Gets the name of this Integration.  # noqa: E501

        descriptive label and unique identifier  # noqa: E501

        :return: The name of this Integration.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Integration.

        descriptive label and unique identifier  # noqa: E501

        :param name: The name of this Integration.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def description(self):
        """Gets the description of this Integration.  # noqa: E501

        longer explanation for the integration  # noqa: E501

        :return: The description of this Integration.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this Integration.

        longer explanation for the integration  # noqa: E501

        :param description: The description of this Integration.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def collections(self):
        """Gets the collections of this Integration.  # noqa: E501

        list of collections that use the integration  # noqa: E501

        :return: The collections of this Integration.  # noqa: E501
        :rtype: list[Collection]
        """
        return self._collections

    @collections.setter
    def collections(self, collections):
        """Sets the collections of this Integration.

        list of collections that use the integration  # noqa: E501

        :param collections: The collections of this Integration.  # noqa: E501
        :type: list[Collection]
        """

        self._collections = collections

    @property
    def created_by(self):
        """Gets the created_by of this Integration.  # noqa: E501

        email of user who created the integration  # noqa: E501

        :return: The created_by of this Integration.  # noqa: E501
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this Integration.

        email of user who created the integration  # noqa: E501

        :param created_by: The created_by of this Integration.  # noqa: E501
        :type: str
        """

        self._created_by = created_by

    @property
    def created_at(self):
        """Gets the created_at of this Integration.  # noqa: E501

        ISO-8601 date  # noqa: E501

        :return: The created_at of this Integration.  # noqa: E501
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Integration.

        ISO-8601 date  # noqa: E501

        :param created_at: The created_at of this Integration.  # noqa: E501
        :type: str
        """

        self._created_at = created_at

    @property
    def s3(self):
        """Gets the s3 of this Integration.  # noqa: E501

        Amazon S3 details, must have one of aws_access_key or aws_role  # noqa: E501

        :return: The s3 of this Integration.  # noqa: E501
        :rtype: S3Integration
        """
        return self._s3

    @s3.setter
    def s3(self, s3):
        """Sets the s3 of this Integration.

        Amazon S3 details, must have one of aws_access_key or aws_role  # noqa: E501

        :param s3: The s3 of this Integration.  # noqa: E501
        :type: S3Integration
        """

        self._s3 = s3

    @property
    def kinesis(self):
        """Gets the kinesis of this Integration.  # noqa: E501

        Amazon Kinesis details, must have one of aws_access_key or aws_role  # noqa: E501

        :return: The kinesis of this Integration.  # noqa: E501
        :rtype: KinesisIntegration
        """
        return self._kinesis

    @kinesis.setter
    def kinesis(self, kinesis):
        """Sets the kinesis of this Integration.

        Amazon Kinesis details, must have one of aws_access_key or aws_role  # noqa: E501

        :param kinesis: The kinesis of this Integration.  # noqa: E501
        :type: KinesisIntegration
        """

        self._kinesis = kinesis

    @property
    def dynamodb(self):
        """Gets the dynamodb of this Integration.  # noqa: E501

        Amazon DynamoDB details, must have one of aws_access_key or aws_role  # noqa: E501

        :return: The dynamodb of this Integration.  # noqa: E501
        :rtype: DynamodbIntegration
        """
        return self._dynamodb

    @dynamodb.setter
    def dynamodb(self, dynamodb):
        """Sets the dynamodb of this Integration.

        Amazon DynamoDB details, must have one of aws_access_key or aws_role  # noqa: E501

        :param dynamodb: The dynamodb of this Integration.  # noqa: E501
        :type: DynamodbIntegration
        """

        self._dynamodb = dynamodb

    @property
    def redshift(self):
        """Gets the redshift of this Integration.  # noqa: E501

        Amazon Redshift details  # noqa: E501

        :return: The redshift of this Integration.  # noqa: E501
        :rtype: RedshiftIntegration
        """
        return self._redshift

    @redshift.setter
    def redshift(self, redshift):
        """Sets the redshift of this Integration.

        Amazon Redshift details  # noqa: E501

        :param redshift: The redshift of this Integration.  # noqa: E501
        :type: RedshiftIntegration
        """

        self._redshift = redshift

    @property
    def gcs(self):
        """Gets the gcs of this Integration.  # noqa: E501

        GCS details  # noqa: E501

        :return: The gcs of this Integration.  # noqa: E501
        :rtype: GcsIntegration
        """
        return self._gcs

    @gcs.setter
    def gcs(self, gcs):
        """Sets the gcs of this Integration.

        GCS details  # noqa: E501

        :param gcs: The gcs of this Integration.  # noqa: E501
        :type: GcsIntegration
        """

        self._gcs = gcs

    @property
    def segment(self):
        """Gets the segment of this Integration.  # noqa: E501

        Segment details  # noqa: E501

        :return: The segment of this Integration.  # noqa: E501
        :rtype: SegmentIntegration
        """
        return self._segment

    @segment.setter
    def segment(self, segment):
        """Sets the segment of this Integration.

        Segment details  # noqa: E501

        :param segment: The segment of this Integration.  # noqa: E501
        :type: SegmentIntegration
        """

        self._segment = segment

    @property
    def kafka(self):
        """Gets the kafka of this Integration.  # noqa: E501

        Kafka details  # noqa: E501

        :return: The kafka of this Integration.  # noqa: E501
        :rtype: KafkaIntegration
        """
        return self._kafka

    @kafka.setter
    def kafka(self, kafka):
        """Sets the kafka of this Integration.

        Kafka details  # noqa: E501

        :param kafka: The kafka of this Integration.  # noqa: E501
        :type: KafkaIntegration
        """

        self._kafka = kafka

    @property
    def mongodb(self):
        """Gets the mongodb of this Integration.  # noqa: E501

        MongoDb details  # noqa: E501

        :return: The mongodb of this Integration.  # noqa: E501
        :rtype: MongoDbIntegration
        """
        return self._mongodb

    @mongodb.setter
    def mongodb(self, mongodb):
        """Sets the mongodb of this Integration.

        MongoDb details  # noqa: E501

        :param mongodb: The mongodb of this Integration.  # noqa: E501
        :type: MongoDbIntegration
        """

        self._mongodb = mongodb

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Integration, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Integration):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, item):
        return getattr(self, item)

    def items(self):
        return self.to_dict().items()

    def __setitem__(self, item, value):
        return seattr(self, item, value)



import re
import yaml
from .command_auth import AuthCommand
from .util.parse_util import parse_collection_path
from datetime import timedelta
from docopt import docopt
from rockset.exception import InputError
from .util.type_util import IntegrationType


class Create(AuthCommand):
    def usage(self):
        return """
Usage:
    rock create [-h | -f=YAML_FILE]
    rock create collection <name> [--description=TEXT] [options] [<data_source_url>...]
    rock create workspace <name> [--description=TEXT]
    rock create integration gcs <name> [--description=TEXT] --service_account_key_file_json=JSON
    rock create integration dynamodb <name> [--description=TEXT] --aws_access_key_id=ID --aws_secret_access_key=SECRET
    rock create integration mongodb <name> [--description=TEXT] --connection_string=CONNECTION_STRING
    rock create integration kinesis <name> [--description=TEXT] ( --aws_role_arn=ARN | --aws_access_key_id=ID --aws_secret_access_key=SECRET )
    rock create integration redshift <name> [--description=TEXT] --aws_access_key_id=ID --aws_secret_access_key=SECRET --host=HOST --port=PORT --username=USERNAME --password=PASSWORD --s3_bucket_path=PATH
    rock create integration s3 <name> [--description=TEXT] ( --aws_role_arn=ARN | --aws_access_key_id=ID --aws_secret_access_key=SECRET )
    rock create integration kafka <name> [--description=TEXT] --kafka_topic_names=TOPIC_1,TOPIC_2,TOPIC_N --kafka_data_format=FORMAT

Options for `rock create`:

    -h, --help                                  Show this help message and exit.

    -f, --file=YAML_FILE                        Create all resources specified in the YAML file.
                                                Run `rock -o yaml describe <collection>` on an
                                                existing collection to see the YAML format.


Options for `rock create collection`:

    --integration=INTEGRATION_NAME              Specify an integration that will be used to access
                                                the source of this collection. For sources that don't need
                                                special access or credentials, this can be left unspecified.

    --retention RETENTION_DURATION              specify the minimum time duration using short-hand notation
                                                (such as 24h, 8d, or 13w) for which documents in this collection
                                                will be retained before being automatically deleted.
                                                (default: none i.e., documents will be retained indefinitely)

    --format CSV                                Specify the format of data in this source. If the data is
                                                comma separted values per line, then specify CSV. Other
                                                formats of data are auto detected and this parameter
                                                should not be specified for non-csv data formats.
                                                All options below only are only applicable when format is CSV.

    --csv-separator ","                         Separator for columns for csv files.

    --csv-encoding "UTF-8"                      Encoding, one of "UTF-8", "UTF-16", "ISO-8859-1".

    --csv-first-line-as-column-names false      If true, then use the first column as the column names.

    --csv-column-names "c1,col2",               A comma separated list of column names

    --csv-column-types "int,boolean",           A comma separated list of column types and should have a
                                                one-to-one mapping to the names specified in column-names.

    --csv-schema-file "/home/schema.yaml"       A file that specifies the schema of the data.
                                                It is an yaml file that specifies the name of each column
                                                and the type of data in each column. If a schema file is
                                                specified, then the values specified by csv-column-names
                                                and csv-column-types are ignored.


Data sources:

    <data_source_url>                           specify the data source to auto ingest in order to
                                                populate the collection, e.g.:
                                                    s3://my-precious-s3-bucket
                                                    s3://my-precious-s3-bucket/data/path/prefix
                                                    dynamodb://my-precious-dynamodb-table-name
                                                    mongodb://my-awesome-mongodb-database/my-precious-mongodb-collection
                                                    kinesis://my-precious-kinesis-stream-name
                                                    gs://server-logs
                                                    gs://server-logs/path/prefix
                                                    kafka://my-precious-kafka-topic-name


Examples:

    Create a collection and source all contents from an AWS S3 bucket:
        $ rock create collection customers s3://customers-mycompany-com

    Create a collection from an AWS S3 bucket but only pull a particular path prefix within the S3 bucket:
        $ rock create collection event-log s3://event-log.mycompany.com/root/path/in/bkt --integration aws-rockset-readonly

    Create a collection with a source that ingests data from a Kinesis stream:
        $ rock create collection events kinesis://click-streams --integration aws-rockset-readonly

    Create a collection with retention set to 10 days:
        $ rock create collection my-event-data --retention="10d"

"""

    def parse_args(self, args):
        parsed = dict(docopt(self.usage(), argv=args, help=False))

        # handle help
        if parsed['--help']:
            ret = self.usage()
            raise SystemExit(ret.strip())

        # see if YAMLFILE was specified
        fn = parsed['--file']
        if fn:
            self.set_batch_items('resource', self._parse_yaml_file(fn))
            return {}

        # process args
        resource = {}

        resource['name'] = parsed['<name>']

        for option in [
            'aws_access_key_id',
            'aws_role_arn',
            'aws_secret_access_key',
            'description',
            'host',
            'connection_string',
            'kafka_data_format',
            'kafka_topic_names',
            'password',
            'port',
            's3_bucket_path',
            'service_account_key_file_json',
            'username',
        ]:
            resource[option] = parsed['--{}'.format(option)]

        if parsed['workspace']:
            resource['type'] = 'WORKSPACE'

        elif parsed['integration']:
            resource['type'] = 'INTEGRATION'
            resource['integration_type'] = IntegrationType.UNKNOWN
            for t in [
                'dynamodb', 'mongodb', 'gcs', 'kinesis', 's3', 'redshift',
                'kafka'
            ]:
                if parsed[t]:
                    resource['integration_type'] = IntegrationType.parse(t)
                    break

        elif parsed['collection']:
            resource['workspace'], resource['name'] = parse_collection_path(
                parsed['<name>']
            )

            sources = []
            integration = None

            integration_name = parsed['--integration']
            if integration_name:
                resource['integration'] = integration_name
                try:
                    integration = self.client.Integration.retrieve(
                        name=integration_name
                    )
                except InputError as e:
                    ret = "Error: {}\n".format(str(e))
                    raise SystemExit(ret.strip())

            # Specifiy the format of data in collection source
            data_format = "auto_detect"
            format_params = None
            if parsed['--format'] is not None:
                data_format = parsed['--format']

            if data_format == "CSV":
                csv_separator = None
                csv_encoding = None
                csv_first_line_as_column_names = None
                csv_column_names = None
                csv_column_types = None
                if parsed['--csv-separator'] is not None:
                    csv_separator = parsed['--csv-separator']
                if parsed['--csv-encoding'] is not None:
                    csv_encoding = parsed['--csv-encoding']
                if parsed['--csv-first-line-as-column-names'] is not None:
                    csv_first_line_as_column_names = parsed[
                        '--csv-first-line-as-column-names']
                if parsed['--csv-column-names'] is not None:
                    csv_column_names_str = parsed['--csv-column-names']
                    csv_column_names = csv_column_names_str.split(',')
                if parsed['--csv-column-types'] is not None:
                    csv_column_types_str = parsed['--csv-column-types']
                    csv_column_types = csv_column_types_str.split(',')

                # if this is a csv schema file, load it and extract
                # the column names and types from it.
                if parsed['--csv-schema-file'] is not None:
                    csv_schema_file = parsed['--csv-schema-file']
                    with open(csv_schema_file, 'r') as stream:
                        data_loaded = yaml.load(stream)

                        # fetch appropriate entry for this collection from the schema file
                        all_columns = data_loaded.get('fields')
                        if all_columns is None:
                            print(
                                'No fields entry ' + resource['name'] +
                                ' in file ' + csv_schema_file
                            )
                            raise SystemExit

                        # one entry in the file shows up as
                        # [{'col1': 'integer'}, {'col2': 'string'}, {'col3': 'boolean'}]

                        csv_column_names = []
                        csv_column_types = []
                        for i in range(len(all_columns)):
                            onecol = all_columns[i]
                            assert len(onecol) == 1
                            colnames = list(all_columns[i].keys())
                            csv_column_names.append(list(onecol.keys())[0])
                            csv_column_types.append(list(onecol.values())[0])

                format_params = self.client.Source.csv_params(
                    separator=csv_separator,
                    encoding=csv_encoding,
                    first_line_as_column_names=csv_first_line_as_column_names,
                    column_names=csv_column_names,
                    column_types=csv_column_types
                )

            resource['type'] = 'COLLECTION'
            for source in parsed['<data_source_url>']:
                if source.startswith('s3://'):
                    sources.append(
                        self._source_s3(source, integration, format_params)
                    )

                elif source.startswith('dynamodb://'):
                    sources.append(self._source_dynamo(source, integration))

                elif source.startswith('mongodb://'):
                    sources.append(self._source_mongo(source, integration))

                elif source.startswith('kinesis://'):
                    sources.append(self._source_kinesis(source, integration))

                elif source.startswith('gs://'):
                    sources.append(
                        self._source_gcs(source, integration, format_params)
                    )

                elif source.startswith('kafka://'):
                    sources.append(self._source_kafka(source, integration))

                else:
                    ret = 'Error: invalid data source URL "{}"\n'.format(source)
                    ret += self.usage()
                    raise SystemExit(ret.strip())

            #handle retention
            if parsed['--retention'] is not None:
                try:
                    retention = self._convert_to_seconds(parsed['--retention'])
                    resource['retention_secs'] = retention
                except ValueError as e:
                    ret = 'Error: invalid argument "{}" for --retention, {}'.format(
                        parsed['--retention'], str(e)
                    )
                    ret += self.usage()
                    raise SystemExit(ret)
                except OverflowError as e:
                    ret = 'Error: invalid value "{}" for --retention, {}'.format(
                        parsed['--retention'], str(e)
                    )
                    ret += self.usage()
                    raise SystemExit(ret)

            resource['sources'] = sources
        else:
            ret = self.usage()
            raise SystemExit(ret.strip())

        return {'resource': resource}

    def go(self):
        self.logger.info('create {}'.format(self.resource))
        rtype = self.resource.pop('type', None)
        if rtype is None:
            return 1
        if rtype == 'COLLECTION':
            return self.go_collection(self.resource)
        elif rtype == 'INTEGRATION':
            return self.go_integration(self.resource)
        elif rtype == 'WORKSPACE':
            return self.go_workspace(self.resource)
        return 1

    def go_collection(self, resource):
        name = resource.pop('name')
        workspace = resource.pop('workspace', 'commons')
        c = self.client.Collection.create(name, workspace=workspace, **resource)
        self.lprint(
            0, 'Collection "%s" was created successfully in workspace "%s".' %
            (c.name, c.workspace)
        )
        return 0

    def go_integration(self, resource):
        integration_type = resource['integration_type']

        try:
            if integration_type == IntegrationType.DYNAMODB:
                self.client.Integration.Dynamodb.create(
                    name=resource['name'],
                    description=resource['description'],
                    aws_access_key_id=resource['aws_access_key_id'],
                    aws_secret_access_key=resource['aws_secret_access_key'],
                )
            elif integration_type == IntegrationType.MONGODB:
                self.client.Integration.MongoDb.create(
                    name=resource['name'],
                    description=resource['description'],
                    connection_uri=resource['connection_string']
                )
            elif integration_type == IntegrationType.GCS:
                self.client.Integration.Gcs.create(
                    name=resource['name'],
                    description=resource['description'],
                    service_account_key_file_json=resource[
                        'service_account_key_file_json'],
                )
            elif integration_type == IntegrationType.KINESIS:
                self.client.Integration.Kinesis.create(
                    name=resource['name'],
                    description=resource['description'],
                    aws_access_key_id=resource['aws_access_key_id'],
                    aws_secret_access_key=resource['aws_secret_access_key'],
                    aws_role_arn=resource['aws_role_arn'],
                )
            elif integration_type == IntegrationType.REDSHIFT:
                self.client.Integration.Redshift.create(
                    name=resource['name'],
                    description=resource['description'],
                    aws_access_key_id=resource['aws_access_key_id'],
                    aws_secret_access_key=resource['aws_secret_access_key'],
                    username=resource['username'],
                    password=resource['password'],
                    host=resource['host'],
                    port=int(resource['port']),
                    s3_bucket_path=resource['s3_bucket_path'],
                )
            elif integration_type == IntegrationType.S3:
                self.client.Integration.S3.create(
                    name=resource['name'],
                    description=resource['description'],
                    aws_access_key_id=resource['aws_access_key_id'],
                    aws_secret_access_key=resource['aws_secret_access_key'],
                    aws_role_arn=resource['aws_role_arn'],
                )
            elif integration_type == IntegrationType.KAFKA:
                self.client.Integration.Kafka.create(
                    name=resource['name'],
                    description=resource['description'],
                    kafka_topic_names=resource['kafka_topic_names'].split(','),
                    kafka_data_format=resource['kafka_data_format'],
                )

            self.lprint(
                0, 'Integration {} was created successfully.'.format(
                    resource['name']
                )
            )

            return 0

        except ValueError as e:
            ret = "Error: {}\n".format(str(e))
            ret += self.usage('integration')
            raise SystemExit(ret.strip())

        return 1

    def go_workspace(self, resource):
        name = resource.pop('name')
        w = self.client.Workspace.create(name, **resource)
        self.lprint(0, 'Workspace "%s" was created successfully.' % (w.name))
        return 0

    def _convert_to_seconds(self, duration):

        num = duration[:-1]
        try:
            num = int(num)
        except ValueError as e:
            ret = 'invalid duration "{}"\n'.format(num)
            raise ValueError(ret)

        unit = duration[-1]
        try:
            if unit == 'h':
                time_delta = timedelta(hours=num)
            elif unit == 'd':
                time_delta = timedelta(days=num)
            elif unit == 'w':
                time_delta = timedelta(weeks=num)
            else:
                ret = 'invalid time unit "{}"\n'.format(unit)
                raise ValueError(ret)
        except OverflowError:
            ret = 'duration "{}" too large for specified time units\n'.format(
                num
            )
            raise OverflowError(ret)

        return int(time_delta.total_seconds())

    def _source_s3(self, s3_url, integration, format_params):
        parts = s3_url[5:].split('/')
        bucket = parts[0]
        path = '/'.join(parts[1:])
        prefix = None
        pattern = None

        matchPattern = re.compile(r'[*?{}]')
        if bool(matchPattern.search(path)):
            pattern = path
        else:
            prefix = path

        return self.client.Source.s3(
            bucket=bucket,
            prefix=prefix,
            pattern=pattern,
            integration=integration,
            format_params=format_params
        )

    def _source_dynamo(self, url, integration):
        table_name = url[11:]

        return self.client.Source.dynamo(
            table_name=table_name, integration=integration
        )

    def _source_mongo(self, url, integration):
        database_and_collection = url[10:]

        first_slash = database_and_collection.find('/')

        # MongoDB databases cannot contain slashes, but collections can.
        # So we need to parse a string like abc/def/ghi as database abc, and collection def/ghi.
        # If you think this is weird: you're right, it is!
        database = database_and_collection[:first_slash]
        collection = database_and_collection[first_slash + 1:]

        return self.client.Source.mongo(
            database_name=database,
            collection_name=collection,
            integration=integration
        )

    def _source_kinesis(self, kinesis_url, integration):
        stream_name = kinesis_url[10:]

        return self.client.Source.kinesis(
            stream_name=stream_name, integration=integration
        )

    def _source_gcs(self, gcs_url, integration, format_params):
        parts = gcs_url[5:].split('/')
        bucket = parts[0]
        prefix = None
        if len(parts) > 1:
            path = '/'.join(parts[1:])
            prefix = path

        return self.client.Source.gcs(
            bucket=bucket,
            prefix=prefix,
            integration=integration,
            format_params=format_params
        )

    def _source_kafka(self, kafka_url, integration):
        kafka_topic_name = kafka_url[8:]

        return self.client.Source.kafka(
            kafka_topic_name=kafka_topic_name, integration=integration
        )



import json
import logging
import sys
import texttable
import yaml

from docopt import docopt
from http.client import HTTPConnection
from rockset.credentials import Credentials
from rockset.exception import (
    AuthError, HTTPError, InputError, LimitReached, NotYetImplemented,
    RequestTimeout, ResourceSuspendedError, ServerError, TransientServerError
)


class Command():
    def _parse_yaml_file(self, fn, upload=False):
        try:
            with open(fn, 'r') as fh:
                resources = self._parse_yaml_filehandle(fh, upload)
        except (IOError, OSError) as e:
            if self.logger.level == logging.DEBUG:
                # Assuming end user doesn't care about our stacktrace
                self.logger.exception(e)
            raise SystemExit(e)
        except ValueError:
            raise ValueError('cannot parse YAML file {}'.format(fn))

        return resources

    def _parse_yaml_filehandle(self, fh, upload):
        # split yaml file at '---' to support multiple resource definitions
        sections = []
        lines = []
        for line in fh:
            if line.rstrip() == '---':
                sections.append(lines)
                lines = []
                continue
            lines.append(line)
        if len(lines) > 0:
            sections.append(lines)

        resources = []
        for section in sections:
            rlist = yaml.load(''.join(section), Loader=yaml.FullLoader)
            if not isinstance(rlist, list):
                rlist = [rlist]
            for ro in rlist:
                if upload:
                    if 'collection' not in ro:
                        raise ValueError(
                            '"collection" is not defined for files '
                            'entry #{}'.format(len(resources))
                        )
                else:
                    if 'name' not in ro:
                        raise ValueError(
                            '"name" is not defined for resource '
                            'entry #{}'.format(len(resources))
                        )
                    if 'type' not in ro:
                        raise ValueError(
                            '"type" is not defined for resource '
                            'entry #{} name={}'.format(
                                len(resources), ro['name']
                            )
                        )
                resources.append(ro)
        return resources

    def __init__(self, *args, **kwargs):
        # set defaults
        self.profile = None
        self.batch_items = None
        self.verbose = 0
        self.logger = logging.getLogger(__name__)

        # override with input
        self.setattrs(*args, **kwargs)

        # initialize loggers
        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter(
                '%(asctime)s %(name)s %(levelname)s - %(message)s',
                datefmt='%H:%M:%S'
            )
        )

        log_level = logging.CRITICAL
        if self.verbose >= 1:
            log_level = logging.WARNING
        if self.verbose >= 2:
            log_level = logging.INFO
        if self.verbose >= 3:
            log_level = logging.DEBUG
            HTTPConnection.debuglevel = 1
        for module in ['rockset', 'requests.packages.urllib3']:
            logger = logging.getLogger(module)
            logger.addHandler(handler)
            logger.setLevel(log_level)
            logger.propagate = True
            self.logger.debug(vars(logger))

        # log some info messages
        self.logger.info('command attrs: {}'.format(vars(self)))

    def setattrs(self, *args, **kwargs):
        for key in kwargs:
            attrkey = key
            if key[:2] == '--':
                attrkey = key[2:]
            elif key[:1] == '-':
                attrkey = key[1:]
            elif key[:1] == '<' and key[-1:] == '>':
                attrkey = key[1:-1]
            setattr(self, attrkey, kwargs[key])

    def set_batch_items(self, field, values):
        self.batch_items = {field: values}

    def get_batch_items(self):
        return self.batch_items

    def iter_batch_items(self):
        if self.batch_items is None:
            return
        field, values = self.batch_items.popitem()
        for v in values:
            yield (field, v)

    def _apicall(self, method, *args, **kwargs):
        self.logger.info('calling %s()' % method)
        self.logger.info('  with args: %s' % str(args))
        self.logger.info('  with kwargs: %s' % str(kwargs))
        result = self.client._apicall(method, *args, **kwargs)
        self.logger.info('received result: %s' % str(result))
        return result

    def parse_args(self, args):
        return dict(docopt(self.usage(), argv=args))

    def validate_args(self, pargs):
        # TODO (veeve): use schema module for validation
        return (pargs is not None)

    def main(self, args):
        # parse input args
        parsed_args = self.parse_args(args)

        # validate input args, some args are not compatible with others
        if not self.validate_args(parsed_args):
            raise SystemExit(self.usage().strip())

        # all is well, set command args to object
        assert (type(parsed_args) == dict)
        self.setattrs(**parsed_args)

        # run through all resources in batch or solo mode
        if self.get_batch_items() is not None:
            return self.main_batch_go()

        # regular mode; go
        return self.main_go()

    def main_batch_go(self):
        # accummulate exceptions in batch mode, don't bail in the middle
        ret = 0
        exceptions = []
        for f, v in self.iter_batch_items():
            try:
                setattr(self, f, v)
                ret += self.main_go()
            except AssertionError as e:
                # unless it is a test failure
                raise
            except SystemExit as e:
                # unless it is a test failure
                raise
            except Exception as e:
                self.logger.exception(e)
                ret = 1
                exceptions.append(e)
                if self.verbose:
                    raise
        return ret

    def main_go(self, *args, **kwargs):
        try:
            ret = self.go(*args, **kwargs)
        except AuthError as e:
            self.logger.error('Error: {}'.format(str(e)))
            self.logger.exception(e)
            self.eprint(
                0, 'Hint: Use "rock configure" to update your '
                'credentials.'
            )
            ret = 126
            if self.verbose:
                raise
        except (
            HTTPError, InputError, LimitReached, NotImplementedError,
            NotYetImplemented, RequestTimeout, ResourceSuspendedError,
            ServerError, TransientServerError
        ) as e:
            self.logger.error('Error: {} {}'.format(type(e).__name__, str(e)))
            self.logger.exception(e)
            self.eprint(0, 'Error: {}'.format(str(e)))
            ret = 1
            if self.verbose:
                raise
        except Exception as e:
            self.logger.error('Unhandled exception: {}'.format(str(e)))
            self.logger.exception(e)
            self.eprint(0, 'Error: {} {}'.format(type(e).__name__, str(e)))
            ret = 1
            if self.verbose:
                raise
        return ret

    def go(self):
        raise NotImplementedError(
            'command "{}" not yet implemented'.format(self.command)
        )

    def require_creds(self):
        creds = Credentials()
        if self.profile is None:
            self.profile = creds.active_profile()
        profile = creds.get(self.profile)
        if ('api_key' not in profile):
            raise AuthError(message='no credentials setup')
        self.api_key = profile['api_key'].strip()
        self.api_key_masked = self.api_key[:4] + '*' * 10 + self.api_key[-4:]
        self.api_server = profile.get('api_server', 'api.rockset.io')
        self.api_server = self.api_server.strip()
        if (
            self.api_server[:7] != 'http://' and
            self.api_server[:8] != 'https://'
        ):
            self.api_server = 'https://{}'.format(self.api_server)
        return (creds, profile)

    def read_stdin(self, what='input'):
        # read from STDIN
        self.lprint(0, 'Enter {} <press ctrl-D to end>:'.format(what))
        lines = []
        while True:
            try:
                line = input("")
            except EOFError:
                break
            lines.append(line)
        return '\n'.join(lines)

    def lprint(self, level, *args, **kwargs):
        if level <= self.verbose:
            print(*args, flush=True, **kwargs)

    def eprint(self, level, *args, **kwargs):
        self.lprint(level, *args, file=sys.stderr, **kwargs)

    def wprint(self, warnings):
        for warning in warnings:
            self.lprint(0, 'Warning: %s' % warning, file=sys.stderr)

    def error(self, *args, **kwargs):
        self.eprint(0, 'Error %s: ' % self.command, end='')
        self.eprint(0, *args, **kwargs)

    def print_list_yaml(self, level, objs):
        self.lprint(level, yaml.dump(objs))
        return len(objs)

    def print_list_json(self, level, objs):
        self.lprint(
            level,
            json.dumps(objs, indent=1, separators=(',', ':'), sort_keys=True)
        )
        return len(objs)

    def print_list_text(self, level, objs, field_order, header):
        # use texttable to print tabular view
        table = texttable.Texttable(max_width=0)
        table.set_deco(
            texttable.Texttable.HEADER | texttable.Texttable.HLINES |
            texttable.Texttable.VLINES
        )

        # setup headers, list fields in field_order first
        # don't assume field_order fields will always exist
        all_fields = set()
        for o in objs:
            if not isinstance(o, dict):
                o = o.to_dict()
            all_fields = all_fields.union(o.keys())

        # assemble fields to display using field_order
        # field_order == '*' refers to all fields that are not
        # listed in field_order
        fields = []
        for f in field_order:
            if f == '*':
                fields += all_fields - set(field_order)
            else:
                fields.append(f)
        if len(fields) == 0:
            fields = sorted(list(all_fields))

        # add header if required
        if header == True:
            table.add_row([f.upper().replace('_', ' ') for f in fields])

        # make everything of type text
        table.set_cols_dtype(['a'] * len(fields))
        table.set_cols_valign(['b'] * len(fields))

        # setup data rows
        for o in objs:
            c_values = []
            if not isinstance(o, dict):
                o = o.to_dict()
            for f in fields:
                v = o.get(f, '')
                # if f is a string, handle Unicode strings
                if isinstance(v, str):
                    v = v.encode('ascii', 'replace')
                # if f is a non-scalar, convert it to unicode safe JSON string
                # json.dumps() is unicode safe, default str(v) is not
                if isinstance(v, (dict, list)):
                    v = json.dumps(v)
                c_values.append(v)
            table.add_row(c_values)

        self.lprint(level, table.draw())
        return len(objs)

    def print_list(
        self, level, objs, field_order=[], header=True, default='text'
    ):
        if objs is None or len(objs) == 0:
            return 0
        if self.format == 'yaml':
            return self.print_list_yaml(level, objs)
        elif self.format == 'json':
            return self.print_list_json(level, objs)
        elif default == 'yaml':
            return self.print_list_yaml(level, objs)
        elif default == 'json':
            return self.print_list_json(level, objs)
        else:
            return self.print_list_text(level, objs, field_order, header)

    def print_one(self, level, one, field_order=[]):
        return self.print_list(level, [one], field_order)

    def prune_empty_fields(self, dictionary):
        result = {}
        for key, val in dictionary.items():
            if not val:
                continue
            elif isinstance(val, dict):
                result[key] = self.prune_empty_fields(val)
            else:
                result[key] = val
        return result



from .dbapi import connect, Connection, Cursor, DictCursor
from .sqlalchemy import RocksetDialect

__all__ = [
    # DBAPI
    connect,
    Connection,
    Cursor,
    DictCursor,

    # SQL Alchemy
    RocksetDialect,
]



"""
Usage
-----
Workspace objects repesents a container of other workspaces and
Rockset collections.

Example
-------
::

    from rockset import Client, Q, F

    # connect securely to Rockset
    rs = Client()

    # create a workspace
    rs.Workspace.create('marketing')

    # create a collection in the workspace
    user_events = rs.Collection.create('user_events', workspace='marketing')

.. _Workspace.create:

Create a new workspace
-----------------------
Creating a workspace using the Client_ object is as simple as
calling ``client.Workspace.create("my-new-workspace")``::

    from rockset import Client
    rs = Client()
    new_ws = rs.Workspace.create("my-new-workspace")

.. _Workspace.list:

List all workspaces
--------------------
List all workspaces using the Client_ object using::

    from rockset import Client
    rs = Client()
    workspaces = rs.Workspace.list()

.. _Workspace.retrieve:

Retrieve an existing workspace
-------------------------------
Retrive a workspace to run various operations on that workspace::

    from rockset import Client
    rs = Client()
    marketing = rs.retrieve('marketing')

.. _Workpace.drop:

Drop a workspace
-----------------
Use the ``drop()`` method to remove a workspace permanently from Rockset.

.. note:: This is a permanent and non-recoverable operation. Beware.

::

    from rockset import Client
    rs = Client()
    marketing = rs.Workspace.retrieve('marketing')
    marketing.drop()

"""
from rockset.swagger_client.api import (WorkspacesApi)
from rockset.swagger_client.models import (CreateWorkspaceRequest)


class WorkspaceClient(object):
    def __init__(self, client):
        self._api = WorkspacesApi(client)

    def create(self, name, description=None, **kwargs):
        request = CreateWorkspaceRequest(
            name=name,
            description=(description if description else ''),
        )
        return self._api.create(body=request).get('data')

    def list(self):
        return self._api.list().get('data')

    def get(self, name=None, **kwargs):
        return self._api.get(name).get('data')

    def delete(self, name=None, **kwargs):
        return self._api.delete(name).get('data')

    def retrieve(self, name=None, **kwargs):
        return self.get(name)

    def drop(self, name=None, **kwargs):
        return self.delete(name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rockset/query_field_path.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rockset/query_field_path.proto',
  package='rockset',
  syntax='proto3',
  serialized_options=b'\n\020io.rockset.protoP\000',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1erockset/query_field_path.proto\x12\x07rockset\"\x1a\n\x18\x41llIndexesQueryFieldPath\"\xb1\x01\n\x17QueryFieldPathComponent\x12\x0f\n\x05index\x18\x01 \x01(\x05H\x00\x12\r\n\x03key\x18\x02 \x01(\tH\x00\x12\x38\n\x0b\x61ll_indexes\x18\x03 \x01(\x0b\x32!.rockset.AllIndexesQueryFieldPathH\x00\x12.\n\rspecial_field\x18\x04 \x01(\x0e\x32\x15.rockset.SpecialFieldH\x00\x42\x0c\n\npath_union\"F\n\x0eQueryFieldPath\x12\x34\n\ncomponents\x18\x01 \x03(\x0b\x32 .rockset.QueryFieldPathComponent* \n\x0cSpecialField\x12\x10\n\x0cSFIELD_SCORE\x10\x00\x42\x14\n\x10io.rockset.protoP\x00\x62\x06proto3'
)

_SPECIALFIELD = _descriptor.EnumDescriptor(
  name='SpecialField',
  full_name='rockset.SpecialField',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SFIELD_SCORE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=323,
  serialized_end=355,
)
_sym_db.RegisterEnumDescriptor(_SPECIALFIELD)

SpecialField = enum_type_wrapper.EnumTypeWrapper(_SPECIALFIELD)
SFIELD_SCORE = 0



_ALLINDEXESQUERYFIELDPATH = _descriptor.Descriptor(
  name='AllIndexesQueryFieldPath',
  full_name='rockset.AllIndexesQueryFieldPath',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=69,
)


_QUERYFIELDPATHCOMPONENT = _descriptor.Descriptor(
  name='QueryFieldPathComponent',
  full_name='rockset.QueryFieldPathComponent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='rockset.QueryFieldPathComponent.index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='rockset.QueryFieldPathComponent.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_indexes', full_name='rockset.QueryFieldPathComponent.all_indexes', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='special_field', full_name='rockset.QueryFieldPathComponent.special_field', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='path_union', full_name='rockset.QueryFieldPathComponent.path_union',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=72,
  serialized_end=249,
)


_QUERYFIELDPATH = _descriptor.Descriptor(
  name='QueryFieldPath',
  full_name='rockset.QueryFieldPath',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='components', full_name='rockset.QueryFieldPath.components', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=251,
  serialized_end=321,
)

_QUERYFIELDPATHCOMPONENT.fields_by_name['all_indexes'].message_type = _ALLINDEXESQUERYFIELDPATH
_QUERYFIELDPATHCOMPONENT.fields_by_name['special_field'].enum_type = _SPECIALFIELD
_QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union'].fields.append(
  _QUERYFIELDPATHCOMPONENT.fields_by_name['index'])
_QUERYFIELDPATHCOMPONENT.fields_by_name['index'].containing_oneof = _QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union']
_QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union'].fields.append(
  _QUERYFIELDPATHCOMPONENT.fields_by_name['key'])
_QUERYFIELDPATHCOMPONENT.fields_by_name['key'].containing_oneof = _QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union']
_QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union'].fields.append(
  _QUERYFIELDPATHCOMPONENT.fields_by_name['all_indexes'])
_QUERYFIELDPATHCOMPONENT.fields_by_name['all_indexes'].containing_oneof = _QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union']
_QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union'].fields.append(
  _QUERYFIELDPATHCOMPONENT.fields_by_name['special_field'])
_QUERYFIELDPATHCOMPONENT.fields_by_name['special_field'].containing_oneof = _QUERYFIELDPATHCOMPONENT.oneofs_by_name['path_union']
_QUERYFIELDPATH.fields_by_name['components'].message_type = _QUERYFIELDPATHCOMPONENT
DESCRIPTOR.message_types_by_name['AllIndexesQueryFieldPath'] = _ALLINDEXESQUERYFIELDPATH
DESCRIPTOR.message_types_by_name['QueryFieldPathComponent'] = _QUERYFIELDPATHCOMPONENT
DESCRIPTOR.message_types_by_name['QueryFieldPath'] = _QUERYFIELDPATH
DESCRIPTOR.enum_types_by_name['SpecialField'] = _SPECIALFIELD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AllIndexesQueryFieldPath = _reflection.GeneratedProtocolMessageType('AllIndexesQueryFieldPath', (_message.Message,), {
  'DESCRIPTOR' : _ALLINDEXESQUERYFIELDPATH,
  '__module__' : 'rockset.query_field_path_pb2'
  # @@protoc_insertion_point(class_scope:rockset.AllIndexesQueryFieldPath)
  })
_sym_db.RegisterMessage(AllIndexesQueryFieldPath)

QueryFieldPathComponent = _reflection.GeneratedProtocolMessageType('QueryFieldPathComponent', (_message.Message,), {
  'DESCRIPTOR' : _QUERYFIELDPATHCOMPONENT,
  '__module__' : 'rockset.query_field_path_pb2'
  # @@protoc_insertion_point(class_scope:rockset.QueryFieldPathComponent)
  })
_sym_db.RegisterMessage(QueryFieldPathComponent)

QueryFieldPath = _reflection.GeneratedProtocolMessageType('QueryFieldPath', (_message.Message,), {
  'DESCRIPTOR' : _QUERYFIELDPATH,
  '__module__' : 'rockset.query_field_path_pb2'
  # @@protoc_insertion_point(class_scope:rockset.QueryFieldPath)
  })
_sym_db.RegisterMessage(QueryFieldPath)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
